# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library of instructions."""
import collections
import json
import random
import re
import string
from collections.abc import Sequence
from typing import Optional, Union

import langdetect
from absl import logging

from mallm.utils.evaluation import instructions_util

_InstructionArgsDtype = Optional[dict[str, Union[int, str, Sequence[str]]]]

_LANGUAGES = instructions_util.LANGUAGE_CODES

# The relational operation for comparison.
_COMPARISON_RELATION = ("less than", "at least")

# The maximum number of sentences.
_MAX_NUM_SENTENCES = 20

# The number of placeholders.
_NUM_PLACEHOLDERS = 4

# The number of bullet lists.
_NUM_BULLETS = 5

# The options of constrained response.
_CONSTRAINED_RESPONSE_OPTIONS = (
        "My answer is yes.", "My answer is no.", "My answer is maybe.")

# The options of starter keywords.
_STARTER_OPTIONS = ("I would say", "My answer is", "I believe",
                                        "In my opinion", "I think", "I reckon", "I feel",
                                        "From my perspective", "As I see it", "According to me",
                                        "As far as I'm concerned", "To my understanding",
                                        "In my view", "My take on it is", "As per my perception")

# The options of ending keywords.
# TODO(jeffreyzhou) add more ending options
_ENDING_OPTIONS = ("Any other questions?",
                                     "Is there anything else I can help with?")

# The number of highlighted sections.
_NUM_HIGHLIGHTED_SECTIONS = 4

# The section spliter.
_SECTION_SPLITER = ("Section", "SECTION")

# The number of sections.
_NUM_SECTIONS = 5

# The number of paragraphs.
_NUM_PARAGRAPHS = 5

# The postscript marker.
_POSTSCRIPT_MARKER = ("P.S.", "P.P.S")

# The number of keywords.
_NUM_KEYWORDS = 2

# The occurrences of a single keyword.
_KEYWORD_FREQUENCY = 3

# The occurrences of a single letter.
_LETTER_FREQUENCY = 10

# The occurrences of words with all capital letters.
_ALL_CAPITAL_WORD_FREQUENCY = 20

# The number of words in the response.
_NUM_WORDS_LOWER_LIMIT = 100
_NUM_WORDS_UPPER_LIMIT = 500


class Instruction:
    """An instruction template."""

    def __init__(self, instruction_id):
        self.id = instruction_id

    def build_description(self, **kwargs):
        raise NotImplementedError("`build_description` not implemented.")

    def get_instruction_args(self):
        raise NotImplementedError("`get_instruction_args` not implemented.")

    def get_instruction_args_keys(self):
        raise NotImplementedError("`get_instruction_args_keys` not implemented.")

    def check_following(self, value):
        raise NotImplementedError("`check_following` not implemented.")


class ResponseLanguageChecker(Instruction):
    """Check the language of the entire response."""

    @classmethod
    def build_description(cls, *, language=None):
        """Build the instruction description.

        Args:
            language: A string representing the expected language of the response. The
                language has to comply to the 97 types defined in
                `langid.py` (https://pypi.org/project/langid/1.1.5/), which follows
                ISO 639-1 codes (https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes);
                for example, `en` for English, `zh` for Chinese, `fr` for French.

        Returns:
            A string representing the instruction description.
        """
        cls._language = language
        if cls._language is None:
            cls._language = random.choice(list(_LANGUAGES.keys()))
        # TODO(tianjianlu): opens the description generation to more choices.
        cls._description_pattern = (
                f"Your ENTIRE response should be in {language} language, no other " +
                "language is allowed.")
        return cls._description_pattern.format(language=_LANGUAGES[cls._language])

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"language": cls._language}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["language"]

    @classmethod
    def check_following(cls, value):
        """Check if the language of the entire response follows the instruction.

        Args:
            value: A string representing the response.

        Returns:
            True if the language of `value` follows instruction; otherwise False.
        """
        assert isinstance(value, str)

        try:
            return langdetect.detect(value) == cls._language
        except langdetect.LangDetectException as e:
            # Count as instruction is followed.
            logging.error(
                    "Unable to detect language for text %s due to %s", value, e
            )    # refex: disable=pytotw.037
            return True


class NumberOfSentences(Instruction):
    """Check the number of sentences."""

    @classmethod
    def build_description(cls, *, num_sentences=None,
                                                relation=None):
        """Build the instruction description.

        Args:
            num_sentences: An integer specifying the number of sentences as a
                threshold.
            relation: A string in (`less than`, `at least`), defining the relational
                operator for comparison.
                Two relational comparisons are supported for now:
                if 'less than', the actual number of sentences < the threshold;
                if 'at least', the actual number of sentences >= the threshold.

        Returns:
            A string representing the instruction description.
        """
        # The number of sentences as a threshold for comparison.
        cls._num_sentences_threshold = num_sentences
        if (cls._num_sentences_threshold is None or
                cls._num_sentences_threshold < 0):
            cls._num_sentences_threshold = random.randint(1, _MAX_NUM_SENTENCES)

        if relation is None:
            cls._comparison_relation = random.choice(_COMPARISON_RELATION)
        elif relation not in _COMPARISON_RELATION:
            raise ValueError("The supported relation for comparison must be in "
                                             f"{_COMPARISON_RELATION}, but {relation} is given.")
        else:
            cls._comparison_relation = relation

        cls._description_pattern = (
                f"Your response should contain {relation} {num_sentences} sentences.")
        return cls._description_pattern.format(
                relation=cls._comparison_relation,
                num_sentences=cls._num_sentences_threshold)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"num_sentences": cls._num_sentences_threshold,
                        "relation": cls._comparison_relation}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["num_sentences", "relation"]

    @classmethod
    def check_following(cls, value):
        """Check if the number of sentences follows the instruction.

        Args:
            value: A string representing the response.

        Returns:
            True if the response follows the instruction.

        Raise:
                ValueError if the string in `instruction_args` is not in
                [`less_than`, `at_least`].
        """
        num_sentences = instructions_util.count_sentences(value)
        if cls._comparison_relation == _COMPARISON_RELATION[0]:
            return num_sentences < cls._num_sentences_threshold
        if cls._comparison_relation == _COMPARISON_RELATION[1]:
            return num_sentences >= cls._num_sentences_threshold
        return None


class PlaceholderChecker(Instruction):
    """Check the placeholders in template writing."""

    @classmethod
    def build_description(cls, *, num_placeholders=None):
        """Build the instruction description.

        Args:
            num_placeholders: An integer denoting the minimum number of
                placeholders required in the response.

        Returns:
            A string representing the instruction description.
        """
        cls._num_placeholders = num_placeholders
        if cls._num_placeholders is None or cls._num_placeholders < 0:
            cls._num_placeholders = random.randint(1, _NUM_PLACEHOLDERS)
        cls._description_pattern = (
                f"The response must contain at least {num_placeholders} placeholders " +
                "represented by square brackets, such as [address].")
        return cls._description_pattern.format(
                num_placeholders=cls._num_placeholders)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"num_placeholders": cls._num_placeholders}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["num_placeholders"]

    @classmethod
    def check_following(cls, value):
        """Check if the number of placeholders follows the instruction.

        Args:
            value: A string representing the response.

        Returns:
            True if the actual number of placeholders in the response is greater than
            or equal to `num_placeholders`; otherwise, False.
        """
        placeholders = re.findall(r"\[.*?\]", value)
        num_placeholders = len(placeholders)
        return num_placeholders >= cls._num_placeholders


class BulletListChecker(Instruction):
    """Checks the bullet list in the prompt."""

    @classmethod
    def build_description(cls, *, num_bullets=None):
        """Build the instruction description.

        Args:
            num_bullets: An integer specifying the exact number of bullet lists
                that is required to appear in the response.

        Returns:
            A string representing the instruction description.
        """
        cls._num_bullets = num_bullets
        if cls._num_bullets is None or cls._num_bullets < 0:
            cls._num_bullets = random.randint(1, _NUM_BULLETS)
        cls._description_pattern = (
                f"Your answer must contain exactly {num_bullets} bullet points. " +
                "Use the markdown bullet points such as:\n" +
                "* This is point 1. \n" +
                "* This is point 2")
        return cls._description_pattern.format(
                num_bullets=cls._num_bullets)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"num_bullets": cls._num_bullets}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["num_bullets"]

    @classmethod
    def check_following(cls, value):
        r"""Check if the number of bullet lists meets the requirement.

        Args:
            value: A string representing the response. The response is expected to
                contain some bullet lists that start with `\*`.

        Returns:
            True if the actual number of bullet lists in the response meets the
            requirement.
        """
        bullet_lists = re.findall(r"^\s*\*[^\*].*$", value, flags=re.MULTILINE)
        bullet_lists_2 = re.findall(r"^\s*-.*$", value, flags=re.MULTILINE)
        num_bullet_lists = len(bullet_lists) + len(bullet_lists_2)
        return num_bullet_lists == cls._num_bullets


class ConstrainedResponseChecker(Instruction):
    """Checks the constrained response."""

    @classmethod
    def build_description(cls):
        """Build the instruction description."""
        # A sequence of string(s) representing the options of the expected response.
        cls._constrained_responses = _CONSTRAINED_RESPONSE_OPTIONS
        cls._description_pattern = (
                "Answer with one of the following options: {response_options}")
        return cls._description_pattern.format(
                response_options=cls._constrained_responses)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return []

    @classmethod
    def check_following(cls, value):
        """Checks if the response matches the constrained options.

        Args:
            value: A string representing the response.

        Returns:
            True if the actual response contains one of the options in the constrained
            responses; otherwise False.
        """
        value = value.strip()
        for constrained_response in cls._constrained_responses:
            if constrained_response in value:
                return True
        return False


class ConstrainedStartChecker(Instruction):
    """Checks the response start."""

    @classmethod
    def build_description(cls, *, starter=None):
        """Build the instruction description.

        Args:
            starter: A string representing the keyward that the response should start
                with.

        Returns:
            A string representing the instruction description.
        """
        cls._starter = starter.strip() if isinstance(starter, str) else starter
        if cls._starter is None:
            cls._starter = random.choice(_STARTER_OPTIONS)
        cls._description_pattern = (
                "During the conversation, when it is your turn, " +
                f"please always start with {starter}")
        return cls._description_pattern.format(starter=cls._starter)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"starter": cls._starter}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["starter"]

    @classmethod
    def check_following(cls, value):
        """Checks if the response starts with the constrained keyword or phrase.

        Args:
            value: A string representing the response.

        Returns:
            True if the response starts with the given phrase or keyword that is
            contained in `instruction_args`; otherwise, False.
        """
        response_pattern = r"^\s*" + cls._starter + r".*$"
        response_with_constrained_start = re.search(response_pattern, value,
                                                                                                flags=re.MULTILINE)
        return bool(response_with_constrained_start)


class HighlightSectionChecker(Instruction):
    """Checks the highlighted section."""

    @classmethod
    def build_description(cls, *, num_highlights=None):
        """Build the instruction description.

        Args:
            num_highlights: An integer specifying the minimum number of highlighted
                sections.

        Returns:
            A string representing the instruction description.
        """
        cls._num_highlights = num_highlights
        if cls._num_highlights is None or cls._num_highlights < 0:
            cls._num_highlights = random.randint(1, _NUM_HIGHLIGHTED_SECTIONS)

        cls._description_pattern = (
                f"Highlight at least {num_highlights} sections in your answer with " +
                "markdown, i.e. *highlighted section*.")

        return cls._description_pattern.format(num_highlights=cls._num_highlights)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"num_highlights": cls._num_highlights}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["num_highlights"]

    @classmethod
    def check_following(cls, value):
        """Checks if the number of highlighted sections meets the requirement.

        Args:
            value: a string repesenting the response. The response is expected to
                contain highlighted sections in the format of *highlighted*.

        Returns:
            True if the actual number of highlighted sections in the format of
            *highlighed sections* meets the minimum requirement; otherwise False.
        """
        num_highlights = 0
        highlights = re.findall(r"\*[^\n\*]*\*", value)
        double_highlights = re.findall(r"\*\*[^\n\*]*\*\*", value)
        for highlight in highlights:
            if highlight.strip("*").strip():
                num_highlights += 1
        for highlight in double_highlights:
            if highlight.removeprefix("**").removesuffix("**").strip():
                num_highlights += 1

        return num_highlights >= cls._num_highlights


class SectionChecker(Instruction):
    """Checks the sections."""

    @classmethod
    def build_description(cls, *, section_spliter=None,
                                                num_sections=None):
        """Build the instruction description.

        Args:
            section_spliter: A string represents the section spliter keyword that
                marks a new section, i.e., `Section` or `SECTION`.
            num_sections: An integer specifying the number of sections.

        Returns:
            A string representing the instruction description.
        """
        cls._section_spliter = section_spliter.strip() if isinstance(
                section_spliter, str) else section_spliter
        if cls._section_spliter is None:
            cls._section_spliter = random.choice(_SECTION_SPLITER)

        cls._num_sections = num_sections
        if cls._num_sections is None or cls._num_sections < 0:
            cls._num_sections = random.randint(1, _NUM_SECTIONS)

        cls._description_pattern = (
                f"Your response must have {num_sections} sections. Mark the beginning " +
                f"of each section with {section_spliter} X, such as:\n" +
                f"{section_spliter} 1\n" +
                "[content of section 1]\n" +
                f"{section_spliter} 2\n" +
                "[content of section 2]")

        return cls._description_pattern.format(
                num_sections=cls._num_sections,
                section_spliter=cls._section_spliter)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"section_spliter": cls._section_spliter,
                        "num_sections": cls._num_sections}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["section_spliter", "num_sections"]

    @classmethod
    def check_following(cls, value):
        """Checks the response contains multiple sections.

        Args:
            value: A string representing the response. The response is expected
                to contain multiple sections (number of sections is greater than 1).
                A new section starts with `Section 1`, where the number denotes the
                section index.

        Returns:
            True if the number of sections in the response is greater than or equal to
            the minimum number of sections; otherwise, False.
        """
        section_splitter_patten = r"\s?" + cls._section_spliter + r"\s?\d+\s?"
        sections = re.split(section_splitter_patten, value)
        num_sections = len(sections) - 1
        return num_sections >= cls._num_sections


class ParagraphChecker(Instruction):
    """Checks the paragraphs."""

    @classmethod
    def build_description(cls, *, num_paragraphs=None):
        """Build the instruction description.

        Args:
            num_paragraphs: An integer specifying the number of paragraphs.

        Returns:
            A string representing the instruction description.
        """
        cls._num_paragraphs = num_paragraphs
        if cls._num_paragraphs is None or cls._num_paragraphs < 0:
            cls._num_paragraphs = random.randint(1, _NUM_PARAGRAPHS)

        cls._description_pattern = (
                f"There should be {num_paragraphs} paragraphs. " +
                "Paragraphs are separated with the markdown divider: ***")

        return cls._description_pattern.format(num_paragraphs=cls._num_paragraphs)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"num_paragraphs": cls._num_paragraphs}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["num_paragraphs"]

    @classmethod
    def check_following(cls, value):
        """Checks the response contains required number of paragraphs.

        Args:
            value: A string representing the response. The response may contain
                paragraphs that are separated by the markdown divider: `***`.

        Returns:
            True if the actual number of paragraphs is the same as required;
            otherwise, False.
        """
        paragraphs = re.split(r"\s?\*\*\*\s?", value)
        num_paragraphs = len(paragraphs)

        for index, paragraph in enumerate(paragraphs):
            if not paragraph.strip():
                if index in {0, len(paragraphs) - 1}:
                    num_paragraphs -= 1
                else:
                    return False

        return num_paragraphs == cls._num_paragraphs


class PostscriptChecker(Instruction):
    """Checks the postscript."""

    @classmethod
    def build_description(cls, *, postscript_marker=None
                                                ):
        """Build the instruction description.

        Args:
            postscript_marker: A string containing the keyword that marks the start
                of the postscript section.

        Returns:
            A string representing the instruction description.
        """
        cls._postscript_marker = postscript_marker.strip() if isinstance(
                postscript_marker, str) else postscript_marker
        if cls._postscript_marker is None:
            cls._postscript_marker = random.choice(_POSTSCRIPT_MARKER)

        cls._description_pattern = (
                "At the end of your response, please explicitly add a postscript " +
                "starting with {postscript}")

        return cls._description_pattern.format(postscript=cls._postscript_marker)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"postscript_marker": cls._postscript_marker}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["postscript_marker"]

    @classmethod
    def check_following(cls, value):
        """Checks if the response follows the postscript format.

        Args:
            value: a string representing the response. The response is expected to
                contain a postscript section.

        Returns:
            True if the response contains a postscript section starting with
            the keyword containing in the `instruction_args`; otherwise False.
        """
        value = value.lower()
        if cls._postscript_marker == "P.P.S":
            postscript_pattern = r"\s*p\.\s?p\.\s?s.*$"
        elif cls._postscript_marker == "P.S.":
            postscript_pattern = r"\s*p\.\s?s\..*$"
        else:
            postscript_pattern = r"\s*" + cls._postscript_marker.lower() + r".*$"
        postscript = re.findall(postscript_pattern, value, flags=re.MULTILINE)
        return bool(postscript)


class RephraseChecker(Instruction):
    """Checks the repharse."""

    @classmethod
    def build_description(cls, *, original_message):
        """Build the instruction description.

        Args:
            original_message: A string representing the original message. The
                rephrased response should only change its words/sentences in between
                its two asterisks, for example, *change me*. Both original and rephrased
                messages should contain the changes in the form of *change me*.

        Returns:
            A string representing the instruction description.
        """
        if not cls.is_change(original_message):
            raise ValueError(f"Message {original_message} does not contain changes "
                                             "in the form of *change me*.")

        cls._reference_without_change = original_message
        cls._description = ("Rephrasing: Your rephrased response should only" +
                                                 "change the words/sentences in between two asterisks" +
                                                 "such as *change me*.")
        return cls._description

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"original_message": cls._reference_without_change}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["original_message"]

    @classmethod
    def check_following(cls, value):
        r"""Checks if the rephrasing follows the instruction.

        Args:
            value: A string representing the response, which is expected to rephras
                the string of `instruction_args`.

        Returns:
            True if `value` and `instruction_args` only differ by the words/sentences
            in between two asterisks such as *change me*; otherwise, False.
        """

        if not cls.is_change(value):
            raise ValueError(f"value {value} does not contain "
                                             "changes in the form of *change me*.")

        response_without_changes = cls.strip_changes(value)
        reference_without_changes = cls.strip_changes(
                cls._reference_without_change)

        return response_without_changes == reference_without_changes

    @classmethod
    def is_change(cls, response):
        """Check if there is change in the response in the form of *change me*."""
        return re.search(r"\*.*\*", response)

    @classmethod
    def strip_changes(cls, response):
        """Strips off the changes."""
        return re.sub(r"\*.*\*", "", response)


class KeywordChecker(Instruction):
    """Check the exisitence of certain keywords."""

    @classmethod
    def build_description(cls, *, keywords=None
                                                ):
        """Build the instruction description.

        Args:
            keywords: A sequence of strings representing the keywords that are
                expected in the response.

        Returns:
            A string representing the instruction description.
        """

        if not keywords:
            cls._keywords = instructions_util.generate_keywords(
                    num_keywords=_NUM_KEYWORDS)
        else:
            cls._keywords = keywords
        cls._keywords = sorted(cls._keywords)

        cls._description_pattern = (f"Include keywords {keywords} in the response.")

        return cls._description_pattern.format(keywords=cls._keywords)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"keywords": cls._keywords}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["keywords"]

    @classmethod
    def check_following(cls, value):
        """Check if the response contain the expected keywords."""
        for keyword in cls._keywords:
            if not re.search(keyword, value, flags=re.IGNORECASE):
                return False
        return True


class KeywordFrequencyChecker(Instruction):
    """Check the keyword frequency."""

    @classmethod
    def build_description(cls, *, keyword=None,
                                                frequency=None,
                                                relation=None):
        """Build the instruction description.

        Args:
            keyword: A string representing a keyword that is expected in the response.
            frequency: An integer specifying the number of times `keyword` is expected
                to appear in the response.
            relation: A string in (`less than`, `at least`), defining the relational
                operator for comparison.
                Two relational comparisons are supported for now:
                if 'less than', the actual number of occurrences < frequency;
                if 'at least', the actual number of occurrences >= frequency.

        Returns:
            A string representing the instruction description.
        """
        if not keyword:
            cls._keyword = instructions_util.generate_keywords(num_keywords=1)[0]
        else:
            cls._keyword = keyword.strip()

        cls._frequency = frequency
        if cls._frequency is None or cls._frequency < 0:
            cls._frequency = random.randint(1, _KEYWORD_FREQUENCY)

        if relation is None:
            cls._comparison_relation = random.choice(_COMPARISON_RELATION)
        elif relation not in _COMPARISON_RELATION:
            raise ValueError("The supported relation for comparison must be in "
                                             f"{_COMPARISON_RELATION}, but {relation} is given.")
        else:
            cls._comparison_relation = relation

        cls._description_pattern = (
                f"In your response, the word {keyword} should appear {relation} " +
                f"{frequency} times.")

        return cls._description_pattern.format(
                keyword=cls._keyword,
                relation=cls._comparison_relation,
                frequency=cls._frequency)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"keyword": cls._keyword,
                        "frequency": cls._frequency,
                        "relation": cls._comparison_relation}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["keyword", "frequency", "relation"]

    @classmethod
    def check_following(cls, value):
        """Checks if the response contain the keyword with required frequency."""
        actual_occurrences = len(re.findall(
                cls._keyword, value, flags=re.IGNORECASE))

        if cls._comparison_relation == _COMPARISON_RELATION[0]:
            return actual_occurrences < cls._frequency
        if cls._comparison_relation == _COMPARISON_RELATION[1]:
            return actual_occurrences >= cls._frequency
        return None


class NumberOfWords(Instruction):
    """Checks the number of words."""

    @classmethod
    def build_description(cls, *, num_words=None,
                                                relation=None):
        """Build the instruction description.

        Args:
            num_words: An integer specifying the number of words contained in the
                response.
            relation: A string in (`less than`, `at least`), defining the relational
                operator for comparison.
                Two relational comparisons are supported for now:
                if 'less than', the actual number of words < num_words;
                if 'at least', the actual number of words >= num_words.

        Returns:
            A string representing the instruction description.
        """

        cls._num_words = num_words
        if cls._num_words is None or cls._num_words < 0:
            cls._num_words = random.randint(
                    _NUM_WORDS_LOWER_LIMIT, _NUM_WORDS_UPPER_LIMIT
            )

        if relation is None:
            cls._comparison_relation = random.choice(_COMPARISON_RELATION)
        elif relation not in _COMPARISON_RELATION:
            raise ValueError("The supported relation for comparison must be in "
                                             f"{_COMPARISON_RELATION}, but {relation} is given.")
        else:
            cls._comparison_relation = relation

        cls._description_pattern = (
                f"Answer with {relation} {num_words} words.")

        return cls._description_pattern.format(
                relation=cls._comparison_relation,
                num_words=cls._num_words)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"num_words": cls._num_words,
                        "relation": cls._comparison_relation}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["num_words", "relation"]

    @classmethod
    def check_following(cls, value):
        """Checks if the response contains the expected number of words."""
        num_words = instructions_util.count_words(value)

        if cls._comparison_relation == _COMPARISON_RELATION[0]:
            return num_words < cls._num_words
        if cls._comparison_relation == _COMPARISON_RELATION[1]:
            return num_words >= cls._num_words
        return None


class JsonFormat(Instruction):
    """Check the Json format."""

    @classmethod
    def build_description(cls):
        cls._description_pattern = (
                "Entire output should be wrapped in JSON format. You can use markdown"
                " ticks such as ```."
        )
        return cls._description_pattern

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return []

    @classmethod
    def check_following(cls, value):
        value = (
                value.strip()
                .removeprefix("```json")
                .removeprefix("```Json")
                .removeprefix("```JSON")
                .removeprefix("```")
                .removesuffix("```")
                .strip()
        )
        try:
            json.loads(value)
        except ValueError as _:
            return False
        return True


class ParagraphFirstWordCheck(Instruction):
    """Check the paragraph and the first word of the nth paragraph."""

    @classmethod
    def build_description(cls, num_paragraphs=None,
                                                nth_paragraph=None,
                                                first_word=None):
        r"""Build the instruction description.

        Args:
            num_paragraphs: An integer indicating the number of paragraphs expected
                in the response. A paragraph is a subset of the string that is
                expected to be separated by '\n\n'.
            nth_paragraph: An integer indicating the paragraph number that we look at.
                Note that n starts from 1.
            first_word: A string that represent the first word of the bth paragraph.

        Returns:
            A string representing the instruction description.
        """
        cls._num_paragraphs = num_paragraphs
        if cls._num_paragraphs is None or cls._num_paragraphs < 0:
            cls._num_paragraphs = random.randint(1, _NUM_PARAGRAPHS)

        cls._nth_paragraph = nth_paragraph
        if (
                cls._nth_paragraph is None
                or cls._nth_paragraph <= 0
                or cls._nth_paragraph > cls._num_paragraphs
        ):
            cls._nth_paragraph = random.randint(1, cls._num_paragraphs + 1)

        cls._first_word = first_word
        if cls._first_word is None:
            cls._first_word = instructions_util.generate_keywords(num_keywords=1)[0]
        cls._first_word = cls._first_word.lower()

        cls._description_pattern = (
                f"There should be {num_paragraphs} paragraphs. " +
                "Paragraphs and only paragraphs are separated with each other by two " +
                "new lines as if it was '\\n\\n' in python. " +
                f"Paragraph {nth_paragraph} must start with word {first_word}.")

        return cls._description_pattern.format(
                num_paragraphs=cls._num_paragraphs,
                nth_paragraph=cls._nth_paragraph,
                first_word=cls._first_word)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"num_paragraphs": cls._num_paragraphs,
                        "nth_paragraph": cls._nth_paragraph,
                        "first_word": cls._first_word}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["num_paragraphs", "nth_paragraph", "first_word"]

    @classmethod
    def check_following(cls, value):
        """Checks for required number of paragraphs and correct first word.

        Args:
            value: a string representing the response. The response may contain
                paragraphs that are separated by two new lines and the first word of
                the nth paragraph will have to match a specified word.

        Returns:
            True if the number of paragraphs is the same as required and the first
            word of the specified paragraph is the same as required. Otherwise, false.
        """

        paragraphs = re.split(r"\n\n", value)
        num_paragraphs = len(paragraphs)

        for paragraph in paragraphs:
            if not paragraph.strip():
                num_paragraphs -= 1

        # check that index doesn't go out of bounds
        if cls._nth_paragraph <= num_paragraphs:
            paragraph = paragraphs[cls._nth_paragraph - 1].strip()
            if not paragraph:
                return False
        else:
            return False

        first_word = ""
        punctuation = {".", ",", "?", "!", "'", '"'}

        # get first word and remove punctuation
        word = paragraph.split()[0].strip()
        # TODO(jeffrey): make more complex?
        word = word.lstrip("'")
        word = word.lstrip('"')

        for letter in word:
            if letter in punctuation:
                break
            first_word += letter.lower()

        return (
                num_paragraphs == cls._num_paragraphs
                and first_word == cls._first_word
        )


# TODO(jeffrey) add relation - at least/at most?
class KeySentenceChecker(Instruction):
    """Check the existence of certain key sentences."""

    @classmethod
    def build_description(cls, key_sentences=None,
                                                num_sentences=None):
        """Build the instruction description.

        Args:
            key_sentences: A sequences of strings representing the key sentences that
                are expected in the response.
            num_sentences: The number of key sentences that are expected to be seen in
                the response.

        Returns:
            A string representing the instruction description.
        """

        if not key_sentences:
            # TODO(jeffrey) make a generate sentences function? wonderwords package
            cls._key_sentences = {"For now, this is fine."}
        else:
            cls._key_sentences = key_sentences

        if not num_sentences:
            cls._num_sentences = random.randint(1, len(cls._key_sentences))
        else:
            cls._num_sentences = num_sentences

        cls._description_pattern = (
                f"Include {num_sentences} of the following sentences {key_sentences}"
        )

        return cls._description_pattern.format(
                num_sentences=cls._num_sentences, key_sentences=cls._key_sentences
        )

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"num_sentences": cls._num_sentences,
                        "key_sentences": list(cls._key_sentences)}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["num_sentences", "key_sentences"]

    @classmethod
    def check_following(cls, value):
        """Checks if the response contains the expected key sentences."""
        count = 0
        sentences = instructions_util.split_into_sentences(value)
        for sentence in cls._key_sentences:
            if sentence in sentences:
                count += 1

        return count == cls._num_sentences


class ForbiddenWords(Instruction):
    """Checks that specified words are not used in response."""

    @classmethod
    def build_description(cls, forbidden_words=None
                                                ):
        """Build the instruction description.

        Args:
            forbidden_words: A sequences of strings respresenting words that are not
                allowed in the response.

        Returns:
            A string representing the instruction description.
        """

        if not forbidden_words:
            cls._forbidden_words = instructions_util.generate_keywords(
                    num_keywords=_NUM_KEYWORDS)
        else:
            cls._forbidden_words = list(set(forbidden_words))
        cls._forbidden_words = sorted(cls._forbidden_words)
        cls._description_pattern = (
                f"Do not include keywords {forbidden_words} in the response."
        )

        return cls._description_pattern.format(
                forbidden_words=cls._forbidden_words
        )

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"forbidden_words": cls._forbidden_words}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["forbidden_words"]

    @classmethod
    def check_following(cls, value):
        """Check if the response does not contain the expected keywords."""
        for word in cls._forbidden_words:
            if re.search(r"\b" + word + r"\b", value, flags=re.IGNORECASE):
                return False
        return True


class RephraseParagraph(Instruction):
    """Checks that the paragraph is rephrased."""

    @classmethod
    def build_description(cls, *, original_paragraph, low, high
                                                ):
        """Builds the instruction description.

        Args:
            original_paragraph: A string presenting the original paragraph. The
                rephrases response should have betweeb low-high words in common.
            low: An integer presenting the lower bound of similar words.
            high: An integer representing the upper bound of similar words.

        Returns:
            A string representing the instruction description.
        """
        # TODO(jeffrey) make more encompassing
        cls._original_paragraph = original_paragraph
        cls._low = low
        cls._high = high

        cls._description = ("Rephrase the following paragraph: " +
                                                 f"{original_paragraph}\nYour response should have " +
                                                 f"between {low} and {high} of the same words. " +
                                                 "Words are the same if and only if all of the " +
                                                 "letters, ignoring cases, are the same. For " +
                                                 "example, 'run' is the same as 'Run' but different " +
                                                 "to 'ran'.")

        return cls._description.format(original_paragraph=original_paragraph,
                                                                        low=cls._low, high=cls._high)

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return {"original_paragraph": cls._original_paragraph,
                        "low": cls._low,
                        "high": cls._high}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["original_paragraph", "low", "high"]

    @classmethod
    def check_following(cls, value):
        val_words = re.findall(r"\w+", value.lower())
        original_words = re.findall(r"\w+", cls._original_paragraph.lower())
        similar_words = 0

        dict_val = collections.Counter(val_words)
        dict_original = collections.Counter(original_words)

        for word in dict_original:
            similar_words += min(dict_original[word], dict_val[word])

        return similar_words >= cls._low and similar_words <= cls._high


class TwoResponsesChecker(Instruction):
    """Check that two responses were given."""

    @classmethod
    def build_description(cls):
        """Build the instruction description."""
        cls._description_pattern = (
                "Give two different responses. Responses and only responses should"
                " be separated by 6 asterisk symbols: ******."
        )
        return cls._description_pattern

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyward args of `build_description`."""
        return

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return []

    @classmethod
    def check_following(cls, value):
        """Checks if the response has two different answers.

        Args:
            value: A string representing the response.

        Returns:
            True if two responses are detected and false otherwise.
        """
        valid_responses = []
        responses = value.split("******")
        for index, response in enumerate(responses):
            if not response.strip():
                if index not in {0, len(responses) - 1}:
                    return False
            else:
                valid_responses.append(response)
        return (
                len(valid_responses) == 2
                and valid_responses[0].strip() != valid_responses[1].strip()
        )


class RepeatPromptThenAnswer(Instruction):
    """Checks that Prompt is first repeated then answered."""

    @classmethod
    def build_description(cls, *, prompt_to_repeat=None):
        """Build the instruction description.

        Args:
            prompt_to_repeat: The prompt that is meant to be repeated.

        Returns:
            A string representing the instruction description.
        """
        if not prompt_to_repeat:
            raise ValueError("prompt_to_repeat must be set.")
        cls._prompt_to_repeat = prompt_to_repeat
        cls._description_pattern = (
                "First repeat the request word for word without change,"
                " then give your answer (1. do not say any words or characters"
                " before repeating the request; 2. the request you need to repeat"
                " does not include this sentence)"
        )
        return cls._description_pattern

    @classmethod
    def get_instruction_args(cls):
        return {"prompt_to_repeat": cls._prompt_to_repeat}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["prompt_to_repeat"]

    @classmethod
    def check_following(cls, value):
        return bool(value.strip().lower().startswith(cls._prompt_to_repeat.strip().lower()))


class EndChecker(Instruction):
    """Checks that the prompt ends with a given phrase."""

    @classmethod
    def build_description(cls, *, end_phrase=None):
        """Build the instruction description.

        Args:
            end_phrase: A string representing the phrase the response should end with.

        Returns:
            A string representing the instruction description.
        """
        cls._end_phrase = (
                end_phrase.strip() if isinstance(end_phrase, str) else end_phrase
        )
        if cls._end_phrase is None:
            cls._end_phrase = random.choice(_ENDING_OPTIONS)
        cls._description_pattern = (
                "Finish your response with this exact phrase {ender}. "
                "No other words should follow this phrase.")
        return cls._description_pattern.format(ender=cls._end_phrase)

    @classmethod
    def get_instruction_args(cls):
        return {"end_phrase": cls._end_phrase}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["end_phrase"]

    @classmethod
    def check_following(cls, value):
        """Checks if the response ends with the expected phrase."""
        value = value.strip().strip("\"").lower()
        cls._end_phrase = cls._end_phrase.strip().lower()
        return value.endswith(cls._end_phrase)


class TitleChecker(Instruction):
    """Checks the response for a title."""

    @classmethod
    def build_description(cls):
        """Build the instruction description."""
        cls._description_pattern = (
                "Your answer must contain a title, wrapped in double angular brackets,"
                " such as <<poem of joy>>."
        )
        return cls._description_pattern

    @classmethod
    def get_instruction_args(cls):
        return None

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return []

    @classmethod
    def check_following(cls, value):
        """Checks if the response contains a title."""
        pattern = r"<<[^\n]+>>"
        re_pattern = re.compile(pattern)
        titles = re.findall(re_pattern, value)

        return any(title.lstrip("<").rstrip(">").strip() for title in titles)


class LetterFrequencyChecker(Instruction):
    """Checks letter frequency."""

    @classmethod
    def build_description(cls, *, letter=None,
                                                let_frequency=None,
                                                let_relation=None):
        """Build the instruction description.

        Args:
            letter: A string representing a letter that is expected in the response.
            let_frequency: An integer specifying the number of times `keyword` is
                expected to appear in the response.
            let_relation: A string in (`less than`, `at least`), defining the
                relational operator for comparison. Two relational comparisons are
                supported for now; if 'less than', the actual number of
                occurrences < frequency; if 'at least', the actual number of
                occurrences >= frequency.

        Returns:
            A string representing the instruction description.
        """
        if (
                not letter
                or len(letter) > 1
                or ord(letter.lower()) < 97
                or ord(letter.lower()) > 122
        ):
            cls._letter = random.choice(list(string.ascii_letters))
        else:
            cls._letter = letter.strip()
        cls._letter = cls._letter.lower()

        cls._frequency = let_frequency
        if cls._frequency is None or cls._frequency < 0:
            cls._frequency = random.randint(1, _LETTER_FREQUENCY)

        if let_relation is None:
            cls._comparison_relation = random.choice(_COMPARISON_RELATION)
        elif let_relation not in _COMPARISON_RELATION:
            raise ValueError(
                    "The supported relation for comparison must be in "
                    f"{_COMPARISON_RELATION}, but {let_relation} is given."
            )
        else:
            cls._comparison_relation = let_relation

        cls._description_pattern = (
                f"In your response, the letter {letter} should appear {let_relation}"
                f" {let_frequency} times."
        )

        return cls._description_pattern.format(
                letter=cls._letter,
                let_frequency=cls._frequency,
                let_relation=cls._comparison_relation,
        )

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyword args of build description."""
        return {"letter": cls._letter,
                        "let_frequency": cls._frequency,
                        "let_relation": cls._comparison_relation}

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["letter", "let_frequency", "let_relation"]

    @classmethod
    def check_following(cls, value):
        """Checks that the response contains the letter at the right frequency."""
        value = value.lower()
        letters = collections.Counter(value)

        if cls._comparison_relation == _COMPARISON_RELATION[0]:
            return letters[cls._letter] < cls._frequency
        return letters[cls._letter] >= cls._frequency


class CapitalLettersEnglishChecker(Instruction):
    """Checks that the response is in english and is in all capital letters."""

    @classmethod
    def build_description(cls):
        """Build the instruction description."""
        cls._description_pattern = (
                "Your entire response should be in English, and in all capital letters."
        )
        return cls._description_pattern

    @classmethod
    def get_instruction_args(cls):
        return None

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return []

    @classmethod
    def check_following(cls, value):
        """Checks that the response is in English and in all capital letters."""
        assert isinstance(value, str)

        try:
            return value.isupper() and langdetect.detect(value) == "en"
        except langdetect.LangDetectException as e:
            # Count as instruction is followed.
            logging.error(
                    "Unable to detect language for text %s due to %s", value, e
            )    # refex: disable=pytotw.037
            return True


class LowercaseLettersEnglishChecker(Instruction):
    """Checks that the response is in english and is in all lowercase letters."""

    @classmethod
    def build_description(cls):
        """Build the instruction description."""
        cls._description_pattern = (
                "Your entire response should be in English, and in all lowercase"
                " letters. No capital letters are allowed."
        )
        return cls._description_pattern

    @classmethod
    def get_instruction_args(cls):
        return None

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return []

    @classmethod
    def check_following(cls, value):
        """Checks that the response is in English and in all lowercase letters."""
        assert isinstance(value, str)

        try:
            return value.islower() and langdetect.detect(value) == "en"
        except langdetect.LangDetectException as e:
            # Count as instruction is followed.
            logging.error(
                    "Unable to detect language for text %s due to %s", value, e
            )    # refex: disable=pytotw.037
            return True


class CommaChecker(Instruction):
    """Checks the response for no commas."""

    @classmethod
    def build_description(cls):
        """Build the instruction description."""
        cls._description_pattern = (
                "In your entire response, refrain from the use of any commas."
        )
        return cls._description_pattern

    @classmethod
    def get_instruction_args(cls):
        return None

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return []

    @classmethod
    def check_following(cls, value):
        """Checks that the response does not contain commas."""
        return not re.search(r"\,", value)


class CapitalWordFrequencyChecker(Instruction):
    """Checks frequency of words with all capital letters."""

    @classmethod
    def build_description(
            cls,
            capital_frequency=None,
            capital_relation=None,
    ):
        """Build the instruction description.

        Args:
            capital_frequency: An integer that represents the number of words that
                should be in all capital letters.
            capital_relation: A string that is 'at least' or 'at most' that refers to
                the frequency.

        Returns:
            A string representing the instruction description.
        """
        cls._frequency = capital_frequency
        if cls._frequency is None:
            cls._frequency = random.randint(1, _ALL_CAPITAL_WORD_FREQUENCY)

        cls._comparison_relation = capital_relation
        if capital_relation is None:
            cls._comparison_relation = random.choice(_COMPARISON_RELATION)
        elif capital_relation not in _COMPARISON_RELATION:
            raise ValueError(
                    "The supported relation for comparison must be in "
                    f"{_COMPARISON_RELATION}, but {capital_relation} is given."
            )

        cls._description_pattern = (
                "In your response, words with all capital letters should appear"
                " {relation} {frequency} times."
        )

        return cls._description_pattern.format(
                frequency=cls._frequency, relation=cls._comparison_relation
        )

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyword args of build description."""
        return {
                "capital_frequency": cls._frequency,
                "capital_relation": cls._comparison_relation,
        }

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return ["capital_frequency", "capital_relation"]

    @classmethod
    def check_following(cls, value):
        """Checks the frequency of words with all capital letters."""
        # Hyphenated words will count as one word
        words = instructions_util.nltk.word_tokenize(value)
        capital_words = [word for word in words if word.isupper()]

        capital_words = len(capital_words)

        if cls._comparison_relation == _COMPARISON_RELATION[0]:
            return capital_words < cls._frequency
        return capital_words >= cls._frequency


class QuotationChecker(Instruction):
    """Checks response is wrapped with double quotation marks."""

    def build_description(self):
        """Build the instruction description."""
        self._description_pattern = (
                "Wrap your entire response with double quotation marks."
        )
        return self._description_pattern

    @classmethod
    def get_instruction_args(cls):
        """Returns the keyword args of build description."""
        return

    @classmethod
    def get_instruction_args_keys(cls):
        """Returns the args keys of `build_description`."""
        return []

    @classmethod
    def check_following(cls, value):
        """Checks if the response is wrapped with double quotation marks."""
        value = value.strip()
        return len(value) > 1 and value[0] == '"' and value[-1] == '"'



import dataclasses
import json
import os
from typing import Any, Optional, Union

from mallm.evaluation.metrics.metric import Metric
from mallm.utils.evaluation import instructions_registry


@dataclasses.dataclass
class InputExample:
    key: int
    instruction_id_list: list[str]
    prompt: str
    kwargs: list[dict[str, Optional[Union[str, int]]]]


@dataclasses.dataclass
class OutputExample:
    instruction_id_list: list[str]
    prompt: str
    response: str
    follow_all_instructions: bool
    follow_instruction_list: list[bool]


class IFEval(Metric):
    """
    A class to evaluate generated texts by instruction-following, following https://github.com/google-research/google-research/blob/master/instruction_following_eval/evaluation_main.py
    """

    _name = "IFEval"

    @staticmethod
    def read_prompt_list(input_jsonl_filename: str) -> list[InputExample]:
        """Read inputs from jsonl."""
        inputs = []
        with open(input_jsonl_filename) as f:
            for lel in f:
                example = json.loads(lel)
                inputs.append(
                        InputExample(key=example["key"],
                            instruction_id_list=example["instruction_id_list"],
                            prompt=example["prompt"],
                            kwargs=example["kwargs"]))
        return inputs

    @staticmethod
    def test_instruction_following_strict(
            inp: InputExample,
            prompt_to_response: dict[str, str],
    ) -> OutputExample:
        """Tests response to see if instrutions are followed."""
        response = prompt_to_response[inp.prompt]
        instruction_list = inp.instruction_id_list
        is_following_list = []

        for index, instruction_id in enumerate(instruction_list):
            instruction_cls = instructions_registry.INSTRUCTION_DICT[instruction_id]
            instruction = instruction_cls(instruction_id)    # type: ignore[no-untyped-call]

            instruction.build_description(**inp.kwargs[index])   # type: ignore[no-untyped-call]
            args = instruction.get_instruction_args()    # type: ignore[no-untyped-call]
            if args and "prompt" in args:
                instruction.build_description(prompt=inp.prompt)     # type: ignore[no-untyped-call]

            if response.strip() and instruction.check_following(response):   # type: ignore[no-untyped-call]
                is_following_list.append(True)
            else:
                is_following_list.append(False)

        return OutputExample(
                instruction_id_list=inp.instruction_id_list,
                prompt=inp.prompt,
                response=response,
                follow_all_instructions=all(is_following_list),
                follow_instruction_list=is_following_list,
        )

    @staticmethod
    def test_instruction_following_loose(
            inp: InputExample,
            prompt_to_response: dict[str, str],
    ) -> OutputExample:
        """Tests response for an upper bound for following instructions."""
        response = prompt_to_response[inp.prompt]
        r = response.split("\n")
        response_remove_first = "\n".join(r[1:]).strip()
        response_remove_last = "\n".join(r[:-1]).strip()
        response_remove_both = "\n".join(r[1:-1]).strip()
        revised_response = response.replace("*", "")
        revised_response_remove_first = response_remove_first.replace("*", "")
        revised_response_remove_last = response_remove_last.replace("*", "")
        revised_response_remove_both = response_remove_both.replace("*", "")
        all_responses = [
                response,
                revised_response,
                response_remove_first,
                response_remove_last,
                response_remove_both,
                revised_response_remove_first,
                revised_response_remove_last,
                revised_response_remove_both,
        ]
        instruction_list = inp.instruction_id_list
        is_following_list = []

        for index, instruction_id in enumerate(instruction_list):
            instruction_cls = instructions_registry.INSTRUCTION_DICT[instruction_id]
            instruction = instruction_cls(instruction_id)    # type: ignore[no-untyped-call]

            instruction.build_description(**inp.kwargs[index])   # type: ignore[no-untyped-call]
            args = instruction.get_instruction_args()    # type: ignore[no-untyped-call]
            if args and "prompt" in args:
                instruction.build_description(prompt=inp.prompt)     # type: ignore[no-untyped-call]

            is_following = False
            for r in all_responses:     # type: ignore[assignment]
                if r.strip() and instruction.check_following(r):     # type: ignore[no-untyped-call, attr-defined]
                    is_following = True
                    break

            is_following_list.append(is_following)

        return OutputExample(
                instruction_id_list=inp.instruction_id_list,
                prompt=inp.prompt,
                response=response,
                follow_all_instructions=all(is_following_list),
                follow_instruction_list=is_following_list,
        )

    @staticmethod
    def read_prompt_to_response_dict(input_jsonl_filename: str) -> dict[str, str]:
        """Creates dictionary matching prompt and response."""
        return_dict = {}
        with open(input_jsonl_filename) as f:
            for lel in f:
                example = json.loads(lel)
                return_dict[example["prompt"]] = example["response"]
        return return_dict

    @staticmethod
    def evaluate(generated_text: str, reference_texts: list[str], dataset_id: Optional[str]) -> dict[str, Any]:

        inputs = IFEval().read_prompt_list(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")) + "/utils/evaluation/input_data.jsonl")
        inputs = [input_obj for input_obj in inputs if str(input_obj.key) == str(dataset_id)]
        prompt_to_response = {}
        prompt_to_response[inputs[0].prompt] = generated_text

        # get instruction following results
        outputs_strict = IFEval().test_instruction_following_strict(inputs[0], prompt_to_response)
        outputs_loose = IFEval().test_instruction_following_loose(inputs[0], prompt_to_response)

        return {"follow_all_instructions_strict": outputs_strict.follow_all_instructions,
                "instruction_id_list_strict": outputs_strict.instruction_id_list,
                "follow_instruction_list_strict": outputs_strict.follow_instruction_list,
                "follow_all_instructions_loose": outputs_loose.follow_all_instructions,
                "instruction_id_list_loose": outputs_loose.instruction_id_list,
                "follow_instruction_list_loose": outputs_loose.follow_instruction_list
                }



import logging

from mallm.models.Chat import Chat
from mallm.models.personas.PersonaGenerator import PersonaGenerator
from mallm.utils.types import InputExample

logger = logging.getLogger("mallm")


class NoPersonaGenerator(PersonaGenerator):
    """
    The NoPersonaGenerator class is used to create agents without a persona.
    """

    def __init__(self, llm: Chat):
        self.llm = llm

    @classmethod
    def generate_persona(
        cls,
        task_description: str,
        already_generated_personas: list[dict[str, str]],
        sample: InputExample,
    ) -> dict[str, str]:
        return {
            "role": f"Participant {len(already_generated_personas) + 1}",
            "description": "A participant of the discussion.",
        }



import logging
from collections import Counter
from typing import Any, Optional

from mallm.agents.panelist import Panelist
from mallm.decision_protocol.protocol import DecisionProtocol
from mallm.models.discussion.ResponseGenerator import ResponseGenerator
from mallm.utils.config import Config
from mallm.utils.enums import DecisionAlteration
from mallm.utils.types import Agreement, VotingResult, VotingResultList, WorkerFunctions

logger = logging.getLogger("mallm")


class ApprovalVoting(DecisionProtocol):
    """
    The Approval Voting decision protocol allows panelists to approve any number of solutions after a certain number of turns.
    """

    _name = "approval_voting"

    def __init__(
        self,
        panelists: list[Panelist],
        num_neutral_agents: int,
        worker_functions: WorkerFunctions,
        vote_turn: int = 3,
    ) -> None:
        super().__init__(panelists, num_neutral_agents, worker_functions)
        self.vote_turn = vote_turn

    def make_decision(
        self,
        agreements: list[Agreement],
        turn: int,
        agent_index: int,
        task: str,
        question: str,
        config: Config,
    ) -> tuple[str, bool, list[Agreement], str, Optional[VotingResultList]]:
        if len(agreements) > self.total_agents:
            agreements = agreements[-self.total_agents :]

        if turn < self.vote_turn or agent_index != self.total_agents - 1:
            return "", False, agreements, "", None

        final_answers_with_confidence, voting_process_string = (
            self.generate_final_answers(agreements, question, task)
        )
        decision, final_answer, results, voting_process_string = (
            self.vote_with_alterations(
                final_answers_with_confidence,
                question,
                task,
                voting_process_string,
                self._name,
                ResponseGenerator.generate_approval_voting_prompt,
                config.voting_protocols_with_alterations,
            )
        )
        return (
            final_answer,
            decision,
            agreements,
            voting_process_string,
            results,
        )

    def process_results(
        self,
        all_votes: dict[str, VotingResult],
        alteration: DecisionAlteration,
        final_answers: list[str],
        votes: list[int],
    ) -> dict[str, VotingResult]:
        winners = []
        most_approved = -1
        if votes:
            # Count approvals for each answer
            # Get the most common approval count
            approval_counts = Counter(votes)
            most_common = approval_counts.most_common()
            most_approved = most_common[0][0]
            most_approved_count = most_common[0][1]

            # Check if there are multiple winners with the same vote count
            winners = [
                candidate
                for candidate, count in approval_counts.items()
                if count == most_approved_count
            ]
        if len(winners) == 1:
            all_votes[alteration.value] = VotingResult(
                votes=votes,
                most_voted=most_approved,
                final_answer=final_answers[most_approved],
                agreed=True,
            )
            logger.info(
                f"Most approved answer from agent {self.panelists[most_approved].short_id}"
            )
        else:
            all_votes[alteration.value] = VotingResult(
                votes=votes,
                most_voted=-1,
                final_answer="",
                agreed=False,
            )
            logger.info("There was a tie. Going for another round of voting.")
        return all_votes

    def process_votes(
        self,
        final_answers: list[str],
        panelist: Panelist,
        vote_str: str,
        vote: Any,
        voting_process_string: str,
    ) -> tuple[str, Any, bool, str]:
        success = False
        approval_list = [
            int(a.strip())
            for a in vote_str.split(",")
            if 0 <= int(a.strip()) < len(final_answers)
        ]
        if approval_list:
            vote.extend(approval_list)
            logger.info(
                f"{panelist.persona} approved answers from {[self.panelists[a].persona for a in approval_list]}"
            )
            voting_process_string += f"{panelist.persona} approved answers from {[self.panelists[a].persona for a in approval_list]}\n"
            success = True
        return vote_str, vote, success, voting_process_string



import logging
import os
import sys
from dataclasses import dataclass, field
from typing import Optional

import requests

from mallm.utils.task_instructions import TASK_INSTRUCTIONS

logger = logging.getLogger("mallm")


@dataclass
class Config:
    # DO NOT overwrite these values once assigned.
    input_json_file_path: str
    output_json_file_path: str
    task_instruction_prompt: str = ""
    task_instruction_prompt_template: Optional[str] = None
    endpoint_url: str = "https://api.openai.com/v1"
    model_name: str = "gpt-3.5-turbo"
    api_key: str = "-"
    max_turns: int = 10
    skip_decision_making: bool = False
    discussion_paradigm: str = "memory"
    response_generator: str = "simple"
    decision_protocol: str = "hybrid_consensus"
    visible_turns_in_memory: int = 2
    debate_rounds: int = 2
    concurrent_api_requests: int = 100
    use_baseline: bool = False
    use_chain_of_thought: bool = True
    num_agents: int = 3
    num_neutral_agents: int = 0
    agent_generator: str = "expert"
    agent_generators_list: list[str] = field(default_factory=list[str])
    trust_remote_code: bool = False
    num_samples: Optional[int] = None
    hf_dataset_split: Optional[str] = "test"
    hf_token: Optional[str] = None
    hf_dataset_version: Optional[str] = None
    hf_dataset_input_column: Optional[str] = None
    hf_dataset_reference_column: Optional[str] = None
    hf_dataset_context_column: Optional[str] = None
    use_ablation: bool = False
    shuffle_input_samples: bool = False
    all_agents_generate_first_draft: bool = False
    all_agents_generate_draft: bool = False
    voting_protocols_with_alterations: bool = False
    calculate_persona_diversity: bool = False
    challenge_final_results: bool = False
    judge_intervention: Optional[str] = None
    judge_metric: Optional[str] = None
    judge_endpoint_url: Optional[str] = None
    judge_api_key: str = "-"
    judge_always_intervene: bool = False

    def __post_init__(self) -> None:
        if (
            not self.task_instruction_prompt
            and self.task_instruction_prompt_template in TASK_INSTRUCTIONS
        ):
            self.task_instruction_prompt = TASK_INSTRUCTIONS[
                self.task_instruction_prompt_template
            ]

    def check_config(self) -> None:
        # TODO: make this more robust and conclusive. All arguments should be checked for validity, making the use of MALLM as fool-proof as possible.
        if not self.task_instruction_prompt:
            logger.error(
                "Please provide an instruction using the --task_instruction_prompt argument or a template using --task_instruction_prompt_template."
            )
            sys.exit(1)
        if os.path.isfile(self.input_json_file_path):
            if not self.input_json_file_path.endswith(".json"):
                logger.error("The dataset path does not seem to be a json file.")
                sys.exit(1)
        else:
            headers = {"Authorization": f"Bearer {self.hf_token}"}
            response = requests.get(
                f"https://datasets-server.huggingface.co/is-valid?dataset={self.input_json_file_path}",
                headers=headers,
            )
            if "preview" not in response.json() or not response.json()["preview"]:
                logger.error("The huggingface dataset cannot be loaded.")
                sys.exit(1)

        if not self.output_json_file_path.endswith(".json"):
            logger.error(
                "The output file does not seem to be a json file. Please specify a file path using --output_json_file_path."
            )
            sys.exit(1)

        if "api.openai.com" in self.endpoint_url and self.api_key == "-":
            logger.error(
                "When using the OpenAI API, you need to provide a key with the argument: --api_key=<your key>"
            )
            sys.exit(1)
        if self.judge_intervention and not self.judge_metric:
            logger.error(
                "When using judge intervention, you need to provide a metric with the argument: --judge_metric=<your metric>."
            )
            sys.exit(1)
        if not self.agent_generators_list:
            self.agent_generators_list = [
                self.agent_generator for i in range(self.num_agents)
            ]
        if (
            self.agent_generators_list
            and len(self.agent_generators_list) != self.num_agents
        ):
            logger.warning(
                f"The length of the provided agent generators ({self.agent_generators_list}) does not match the number of agents (3). Setting num_agents={len(self.agent_generators_list)}."
            )
            self.num_agents = len(self.agent_generators_list)
        if self.endpoint_url.endswith("/"):
            logger.warning("Removing trailing / from the endpoint url.")
            self.endpoint_url = self.endpoint_url[:-1]
        try:
            logger.info("Testing availability of the endpoint...")
            page = requests.head(self.endpoint_url.replace("/v1", ""))
            logger.info("Status: " + str(page.status_code))
            assert page.status_code == 200
        except Exception as e:
            logger.error("HTTP Error: Could not connect to the provided endpoint url.")
            logger.error(e)
            sys.exit(1)
        if self.concurrent_api_requests > 250:
            logger.warning(
                "concurrent_api_requests is very large. Please make sure the API endpoint you are using can handle that many simultaneous requests."
            )
        # import here to avoid circular imports
        from mallm.utils.dicts import (  # noqa PLC0415
            DECISION_PROTOCOLS,
            DISCUSSION_PARADIGMS,
            RESPONSE_GENERATORS,
        )

        if self.response_generator not in RESPONSE_GENERATORS:
            logger.error(
                f"Invalid response generator: {self.response_generator}. Available options are: {RESPONSE_GENERATORS.keys()}."
            )
            sys.exit(1)
        if self.discussion_paradigm not in DISCUSSION_PARADIGMS:
            logger.error(
                f"Invalid discussion paradigm: {self.discussion_paradigm}. Available options are: {DISCUSSION_PARADIGMS.keys()}."
            )
            sys.exit(1)
        if self.decision_protocol not in DECISION_PROTOCOLS:
            logger.error(
                f"Invalid decision protocol: {self.decision_protocol}. Available options are: {DECISION_PROTOCOLS.keys()}."
            )
            sys.exit(1)



from typing import Any, Optional

from rouge_score import rouge_scorer

from mallm.evaluation.metrics.metric import Metric


class ROUGE(Metric):
    """
    A class to evaluate the ROUGE score for text generation tasks.
    """

    _name = "ROUGE"

    @staticmethod
    def evaluate(generated_text: str, reference_texts: list[str], dataset_id: Optional[str]) -> dict[str, Any]:
        scorer = rouge_scorer.RougeScorer(
            rouge_types=["rouge1", "rouge2", "rouge3", "rougeL"], use_stemmer=True
        )
        scores = scorer.score(
            target=reference_texts[0], prediction=generated_text
        )  # rouge only takes one reference
        return {
            "rouge1": scores["rouge1"].fmeasure,
            "rouge2": scores["rouge2"].fmeasure,
            "rouge3": scores["rouge3"].fmeasure,
            "rougeL": scores["rougeL"].fmeasure,
        }



import logging
from typing import Optional

from mallm.models.Chat import Chat
from mallm.models.discussion.ResponseGenerator import ResponseGenerator
from mallm.utils.types import Response, TemplateFilling

logger = logging.getLogger("mallm")


class FreeTextResponseGenerator(ResponseGenerator):
    """
    The FreeTextResponseGenerator class is designed to generate free-form text responses based on the input prompts and tasks.
    It utilizes a chat interface to interact with the language model and generate responses that are not limited by specific templates or structures.
    """

    _name = "freetext"

    def __init__(self, llm: Chat, judge_llm: Optional[Chat] = None):
        self.llm = llm
        self.judge_llm = judge_llm
        self.base_prompt = {
            "role": "system",
            "content": "You are participating in a discussion to solve the provided task.",
        }
        self.base_prompt_baseline = {
            "role": "system",
            "content": "Solve the provided task. Do not ask back questions. Clearly indicate your final solution after the text 'Final Solution:'.",
        }

    def generate_baseline(
        self, task_instruction: str, input_str: str, chain_of_thought: bool
    ) -> Response:
        prompt_content = f"""
Task: {task_instruction}
Input: {input_str}
        """  # input has context appended
        prompt = [
            self.base_prompt_baseline,
            {
                "role": "user",
                "content": prompt_content,
            },
        ]
        return self.generate_response(
            prompt, task_instruction, input_str, chain_of_thought, None, True, True
        )

    def generate_response(
        self,
        current_prompt: list[dict[str, str]],
        task_instruction: str,
        input_str: str,
        chain_of_thought: bool,
        agreement: Optional[bool],
        baseline: bool,
        drafting: bool,
        judging: bool = False,
    ) -> Response:
        if chain_of_thought:
            current_prompt.append(
                {
                    "role": "user",
                    "content": "Let's think step by step.",
                }
            )

        retry = 0
        while retry < 10:
            res = self.llm.invoke(current_prompt)

            response = Response(
                agreement=(
                    agreement
                    if agreement is not None
                    else self.extract_agreement(res, drafting)
                ),
                message=res,
                solution=self.extract_result(res, task_instruction, input_str),
            )

            if response.agreement is None and not drafting and not baseline and not judging:
                retry += 1
                continue
            break  # success
        if retry >= 10:
            logger.error(
                f"After 10 retries the response could not be decoded. \nPrompt: {current_prompt} \nResponse string: {response}"
            )
            raise Exception("Could not decode response.")
        return response

    def generate_feedback(
        self, data: TemplateFilling, chain_of_thought: bool
    ) -> Response:
        instr_prompt = {
            "role": "user",
            "content": "Based on the current solution, give constructive feedback. If you agree, answer with [AGREE], else answer with [DISAGREE] and explain why.",
        }
        if not data.current_draft:
            instr_prompt = {
                "role": "user",
                "content": "Give constructive feedback.",
            }
        current_prompt = [
            self.base_prompt,
            *self.get_filled_template(data),
            instr_prompt,
        ]
        return self.generate_response(
            current_prompt,
            data.task_instruction,
            data.input_str,
            chain_of_thought,
            None,
            False,
            False,
        )

    def generate_improve(
        self, data: TemplateFilling, chain_of_thought: bool
    ) -> Response:
        instr_prompt = {
            "role": "user",
            "content": "Improve the current solution. If you agree with the current solution, answer with [AGREE], else answer with [DISAGREE] and explain why and provide an improved solution.",
        }
        if not data.current_draft:
            instr_prompt = {
                "role": "user",
                "content": "Propose a solution.",
            }
        current_prompt = [
            self.base_prompt,
            *self.get_filled_template(data),
            instr_prompt,
        ]
        return self.generate_response(
            current_prompt,
            data.task_instruction,
            data.input_str,
            chain_of_thought,
            None,
            False,
            False,
        )

    def generate_draft(self, data: TemplateFilling, chain_of_thought: bool) -> Response:
        instr_prompt = {
            "role": "user",
            "content": "Based on the provided feedback, carefully re-examine your previous solution. Provide a revised solution.",
        }
        if not data.current_draft:
            instr_prompt = {
                "role": "user",
                "content": "Propose a solution.",
            }
        current_prompt = [
            self.base_prompt,
            *self.get_filled_template(data),
            instr_prompt,
        ]
        return self.generate_response(
            current_prompt,
            data.task_instruction,
            data.input_str,
            chain_of_thought,
            None,
            False,
            True,
        )

    def extract_result(
        self, result: Optional[str], task_instruction: str, input_text: str
    ) -> str:
        current_prompt = self.generate_final_answer_prompt(
            input_sample=input_text,
            task=task_instruction,
            previous_answer=result
        )
        return str(self.llm.invoke(current_prompt))

    def generate_ablation(
        self,
        task_instruction: str,
        input_str: str,
        current_solution: str,
        chain_of_thought: bool,
    ) -> Response:
        prompt_content = f"""
When, faced with a task, improve the current solution.
Task: {task_instruction}
Input: {input_str}
Current solution: {current_solution}
"""  # input has context appended
        prompt = [
            {
                "role": "user",
                "content": prompt_content,
            },
        ]
        return self.generate_response(
            current_prompt=prompt,
            task_instruction=task_instruction,
            input_str=input_str,
            chain_of_thought=chain_of_thought,
            agreement=None,
            baseline=True,
            drafting=True,
        )

    def generate_policy_intervention(self, data: TemplateFilling, provide_labels: bool = True) -> Response:
        logger.error(f"Policy Intervention is not implemented for this response generator. {self.__class__.__name__}")
        raise NotImplementedError("Policy Intervention is not implemented for this response generator.")

    def generate_judgement(self, data: TemplateFilling, answer_before: str, answer_after: str) -> Response:
        logger.error(f"Judgement is not implemented for this response generator. {self.__class__.__name__}")
        raise NotImplementedError("Judgement is not implemented for this response generator.")



import ast
import logging
from typing import Any, Optional

from mallm.agents.panelist import Panelist
from mallm.decision_protocol.protocol import DecisionProtocol
from mallm.models.discussion.ResponseGenerator import ResponseGenerator
from mallm.utils.config import Config
from mallm.utils.enums import DecisionAlteration
from mallm.utils.types import Agreement, VotingResult, VotingResultList, WorkerFunctions

logger = logging.getLogger("mallm")


class CumulativeVoting(DecisionProtocol):
    """
    The Cumulative Voting decision protocol allows panelists to distribute 10 points among the solutions.
    The solution with the highest total points is selected as the final decision.
    """

    _name = "cumulative_voting"

    def __init__(
        self,
        panelists: list[Panelist],
        num_neutral_agents: int,
        worker_functions: WorkerFunctions,
        vote_turn: int = 3,
    ) -> None:
        super().__init__(panelists, num_neutral_agents, worker_functions)
        self.vote_turn = vote_turn

    def make_decision(
        self,
        agreements: list[Agreement],
        turn: int,
        agent_index: int,
        task: str,
        question: str,
        config: Config,
    ) -> tuple[str, bool, list[Agreement], str, Optional[VotingResultList]]:
        if len(agreements) > self.total_agents:
            agreements = agreements[-self.total_agents :]

        if turn < self.vote_turn or agent_index != self.total_agents - 1:
            return "", False, agreements, "", None

        final_answers_with_confidence, voting_process_string = (
            self.generate_final_answers(agreements, question, task)
        )

        decision, final_answer, results, voting_process_string = (
            self.vote_with_alterations(
                final_answers_with_confidence,
                question,
                task,
                voting_process_string,
                self._name,
                ResponseGenerator.generate_cumulative_voting_prompt,
                config.voting_protocols_with_alterations,
            )
        )
        return (
            final_answer,
            decision,
            agreements,
            voting_process_string,
            results,
        )

    def process_results(
        self,
        all_votes: dict[str, VotingResult],
        alteration: DecisionAlteration,
        final_answers: list[str],
        votes: Any,
    ) -> dict[str, VotingResult]:
        # Aggregate points for each solution
        total_points = [0] * len(final_answers)
        for points in votes:
            for index, point in points.items():
                total_points[index] += point
        # Determine the solution with the highest points, break ties by selecting the first solution and go for another round
        max_points = max(total_points)
        best_solution_index = total_points.index(max_points)
        best_answers = [
            final_answers[i]
            for i, score in enumerate(total_points)
            if score == max_points
        ]
        if len(best_answers) == 1:
            all_votes[alteration.value] = VotingResult(
                votes=votes,
                most_voted=best_solution_index,
                final_answer=final_answers[best_solution_index],
                agreed=True,
            )
            logger.info(
                f"Selected answer from agent {self.panelists[best_solution_index].short_id} with {max_points} points"
            )
        else:
            all_votes[alteration.value] = VotingResult(
                votes=votes,
                most_voted=-1,
                final_answer="",
                agreed=False,
            )
            logger.info("There was a tie. Going for another round of voting.")
        return all_votes

    def process_votes(
        self,
        final_answers: list[str],
        panelist: Panelist,
        vote_str: str,
        vote: Any,
        voting_process_string: str,
    ) -> tuple[str, Any, bool, str]:
        success = False
        vote_str = vote_str.replace("\n", "").replace(" ", "").strip()
        points_dict = ast.literal_eval(vote_str)
        points_dict = {int(k): int(v) for k, v in points_dict.items()}
        if self.validate_points_distribution(points_dict, len(final_answers)):
            vote.append(points_dict)
            logger.info(f"{panelist.persona} allocated points: {points_dict}")
            voting_process_string += (
                f"{panelist.persona} allocated points: {points_dict}\n"
            )
            success = True
        return vote_str, vote, success, voting_process_string

    @staticmethod
    def validate_points_distribution(
        points_dict: dict[int, int], num_solutions: int
    ) -> bool:
        total_points = sum(points_dict.values())
        if total_points != 10:
            return False
        for index in points_dict:
            if not isinstance(index, int) or not (0 <= index < num_solutions):
                return False
        return not any(x < 0 for x in points_dict.values())



from typing import Optional


def extract_draft(response: Optional[str]) -> Optional[str]:
    match_str = "final solution"
    if response:
        position = response.lower().rfind(match_str)
        if position == -1:
            return None
        matched_str = response[position + len(match_str) :].strip()
        if matched_str[0] == "]" or matched_str[0] == ":":
            matched_str = matched_str[1:].strip()
        return matched_str.split("\n\n")[
            0
        ]  # because LM tends to add extra explanation afterwards
    return None



from __future__ import annotations

import dataclasses
import logging
import uuid
from typing import TYPE_CHECKING, Optional

import httpx

from mallm.models.Chat import Chat
from mallm.models.discussion.ResponseGenerator import ResponseGenerator

if TYPE_CHECKING:
    from mallm.coordinator import Coordinator

from mallm.utils.types import Agreement, Memory, TemplateFilling

logger = logging.getLogger("mallm")


class Agent:
    """
    Represents an Agent in the discussion, capable of improving, drafting, and providing feedback.
    """

    def __init__(
        self,
        llm: Chat,
        client: httpx.Client,
        coordinator: Coordinator,
        response_generator: ResponseGenerator,
        persona: str,
        persona_description: str,
        chain_of_thought: bool = False,
        drafting_agent: bool = False,
    ):
        """
        Initializes an Agent with the necessary components for facilitating discussions.
        """
        self.id = str(uuid.uuid4())
        self.short_id = self.id[:4]
        self.persona = persona
        self.persona_description = persona_description
        self.coordinator = coordinator
        self.llm = llm
        self.response_generator = response_generator
        self.client = client
        self.chain_of_thought = chain_of_thought
        self.drafting_agent = drafting_agent
        self.memory: dict[str, Memory] = {}
        logger.info(
            f"Creating agent [bold blue]{self.short_id}[/] with personality [bold blue]{self.persona}[/]: {self.persona_description}"
        )

    def improve(
        self,
        unique_id: int,
        turn: int,
        memory_ids: list[int],
        template_filling: TemplateFilling,
        agreements: list[Agreement],
    ) -> tuple[str, Memory, list[Agreement]]:
        """
        Agent improves a solution based on the current state of the discussion.
        """
        logger.debug(f"Agent [bold blue]{self.short_id}[/] is improving the solution.")
        response = self.response_generator.generate_improve(
            template_filling, self.chain_of_thought
        )
        logger.debug(
            f"Agent [bold blue]{self.short_id}[/] {'agreed' if response.agreement else 'disagreed'} with the solution."
        )
        agree = False if unique_id == 0 else response.agreement
        agreements.append(
            Agreement(
                agreement=agree,
                response=response.message,
                solution=response.solution if not agree else agreements[-1].solution,
                agent_id=self.id,
                persona=self.persona,
                message_id=unique_id,
            )
        )

        memory = Memory(
            message_id=unique_id,
            turn=turn,
            agent_id=self.id,
            persona=self.persona,
            contribution="improve",
            message=response.message,
            agreement=agreements[-1].agreement,
            solution=response.solution if not agree else agreements[-2].solution,
            memory_ids=memory_ids,
            additional_args=dataclasses.asdict(template_filling),
        )
        self.coordinator.memory.append(memory)
        return response.message, memory, agreements

    def draft(
        self,
        unique_id: int,
        turn: int,
        memory_ids: list[int],
        template_filling: TemplateFilling,
        agreements: list[Agreement],
        is_neutral: bool = False,
    ) -> tuple[str, Memory, list[Agreement]]:
        """
        Agent drafts a solution based on the current state of the discussion.
        """
        logger.debug(f"Agent [bold blue]{self.short_id}[/] is drafting a solution.")
        response = self.response_generator.generate_draft(
            template_filling, self.chain_of_thought
        )
        agreements.append(
            Agreement(
                agreement=None,
                response=response.message,
                solution=response.solution,
                agent_id=self.id,
                persona=self.persona,
                message_id=unique_id,
            )
        )

        memory = Memory(
            message_id=unique_id,
            turn=turn,
            agent_id=self.id,
            persona=self.persona,
            contribution="draft",
            message=response.message,
            agreement=None,
            solution=response.solution,
            memory_ids=memory_ids,
            additional_args=dataclasses.asdict(template_filling),
        )
        self.coordinator.memory.append(memory)
        return response.message, memory, agreements

    def feedback(
        self,
        unique_id: int,
        turn: int,
        memory_ids: list[int],
        template_filling: TemplateFilling,
        agreements: list[Agreement],
    ) -> tuple[str, Memory, list[Agreement]]:
        """
        Agent provides feedback to a solution based on the current state of the discussion.
        """
        logger.debug(
            f"Agent [bold blue]{self.short_id}[/] provides feedback to a solution."
        )
        response = self.response_generator.generate_feedback(
            template_filling, self.chain_of_thought
        )
        logger.debug(
            f"Agent [bold blue]{self.short_id}[/] {'agreed' if response.agreement else 'disagreed'} with the solution."
        )
        logger.debug(f"Agent [bold blue]{self.short_id}[/]: {response.message}")
        agreements.append(
            Agreement(
                agreement=response.agreement,
                response=response.message,
                solution="",
                agent_id=self.id,
                persona=self.persona,
                message_id=unique_id,
            )
        )

        memory = Memory(
            message_id=unique_id,
            turn=turn,
            agent_id=self.id,
            persona=self.persona,
            contribution="feedback",
            message=response.message,
            agreement=agreements[-1].agreement,
            solution=None,
            memory_ids=memory_ids,
            additional_args=dataclasses.asdict(template_filling),
        )
        self.coordinator.memory.append(memory)
        return response.message, memory, agreements

    def update_memory(self, memory: Memory) -> None:
        """
        Updates the memory of this agent with another discussion entry.
        """
        self.memory[str(memory.message_id)] = memory

    def get_memories(
        self,
        context_length: Optional[int] = None,
        turn: Optional[int] = None,
        include_this_turn: bool = True,
    ) -> tuple[Optional[list[Memory]], list[int], Optional[str]]:
        """
        Retrieves memory data from the agents memory
        """
        memories: list[Memory]
        memory_ids = []
        current_draft = None

        memories = sorted(
            self.memory.values(), key=lambda x: x.message_id, reverse=False
        )
        context_memory = []
        for memory in memories:
            if (
                context_length
                and turn
                and memory.turn >= turn - context_length
                and (turn > memory.turn or include_this_turn)
            ):
                context_memory.append(memory)
                memory_ids.append(int(memory.message_id))
                if memory.contribution == "draft" or (
                    memory.contribution == "improve" and memory.agreement is False
                ):
                    current_draft = memory.solution
            else:
                context_memory.append(memory)
                memory_ids.append(int(memory.message_id))
                if memory.contribution == "draft" or (
                    memory.contribution == "improve" and memory.agreement is False
                ):
                    current_draft = memory.solution

        return context_memory, memory_ids, current_draft

    def forget_memories(self, turn: int) -> None:
        keys_to_delete = [key for key, memory in self.memory.items() if memory.turn == turn]
        for key in keys_to_delete:
            del self.memory[key]
        logger.debug(f"Forgot memories {keys_to_delete} from turn {turn} from agent {self.id}")

    def get_own_messages(self, context_length: Optional[int] = None) -> list[str]:
        """
        Retrieves memory from the agents memory bucket as a string
        context_length refers to the amount of turns the agent can use as rationale
        Returns: list of strings
        """
        memories, _, _ = self.get_memories(context_length=context_length)
        if memories:
            own_messages = [
                memory.message for memory in memories if memory.agent_id == self.id
            ]
        else:
            own_messages = []
        return own_messages

    def get_discussion_history(
        self,
        context_length: Optional[int] = None,
        turn: Optional[int] = None,
        include_this_turn: bool = True,
    ) -> tuple[Optional[list[dict[str, str]]], list[int], Optional[str]]:
        """
        Retrieves memory from the agents memory as a string
        context_length refers to the amount of turns the agent can memorize the previous discussion
        """
        memories, memory_ids, current_draft = self.get_memories(
            context_length=context_length,
            turn=turn,
            include_this_turn=include_this_turn,
        )
        if memories:
            discussion_history = []
            for memory in memories:
                if memory.agent_id == self.id:
                    discussion_history.append(
                        {"role": "assistant", "content": memory.message}
                    )
                else:
                    discussion_history.append(
                        {
                            "role": "user",
                            "content": f"{memory.persona}: {memory.message}",
                        }
                    )
        else:
            discussion_history = None
        return discussion_history, memory_ids, current_draft



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from rich.panel import Panel
from rich.progress import Console
from rich.text import Text

from mallm.agents.draftProposer import DraftProposer
from mallm.agents.judge import Judge
from mallm.agents.panelist import Panelist
from mallm.utils.types import Agreement, Memory, TemplateFilling, VotingResultList

if TYPE_CHECKING:
    from mallm.coordinator import Coordinator
    from mallm.utils.config import Config
logger = logging.getLogger("mallm")


class DiscoursePolicy(ABC):
    def __init__(self, paradigm_str: str = "") -> None:
        self.paradigm_str = paradigm_str
        self.decision = False
        self.turn = 0
        self.unique_id = 0
        self.memories: list[Memory] = []
        self.draft = ""
        self.agreements: list[Agreement] = []

    def discuss(
        self,
        coordinator: Coordinator,
        task_instruction: str,
        input_str: str,
        solution: str,
        config: Config,
        console: Optional[Console] = None,
    ) -> tuple[
        Optional[str], int, list[Agreement], bool, dict[int, Optional[VotingResultList]]
    ]:
        logger.info(self.paradigm_str)
        voting_process_string = ""
        voting_results_per_turn: dict[int, Optional[VotingResultList]] = {}

        if console is None:
            console = Console()
        while (
            not self.decision or config.skip_decision_making
        ) and self.turn < config.max_turns:
            self.turn += 1
            logger.debug(f"Ongoing. Current turn: {self.turn}")

            for i, agent in enumerate(coordinator.agents):
                discussion_history, memory_ids, current_draft = (
                    agent.get_discussion_history(
                        context_length=config.visible_turns_in_memory,
                        turn=self.turn,
                    )
                )
                if (
                    self.turn == 1 and config.all_agents_generate_first_draft
                ) or config.all_agents_generate_draft:
                    current_draft = None
                    discussion_history = None

                template_filling = TemplateFilling(
                    task_instruction=task_instruction,
                    input_str=input_str,
                    current_draft=current_draft,
                    persona=agent.persona,
                    persona_description=agent.persona_description,
                    agent_memory=discussion_history,
                )

                if isinstance(agent, DraftProposer):
                    template_filling.feedback_sentences = None
                    self.draft_proposer_call(
                        draft_proposer=agent,
                        coordinator=coordinator,
                        agent_index=i,
                        memory_ids=memory_ids,
                        template_filling=template_filling,
                    )
                elif isinstance(agent, Panelist):
                    self.panelist_call(
                        agent=agent,
                        coordinator=coordinator,
                        agent_index=i,
                        memory_ids=memory_ids,
                        template_filling=template_filling,
                    )
                elif isinstance(agent, Judge):
                    continue    # executes after decision protocol
                else:
                    logger.error("Agent type not recognized.")
                    raise Exception("Agent type not recognized.")
                self.unique_id += 1
                self.memories = []

                if coordinator.decision_protocol is None:
                    logger.error("No decision protocol module found.")
                    raise Exception("No decision protocol module found.")

                (
                    self.draft,
                    self.decision,
                    self.agreements,
                    voting_process_string,
                    additional_voting_results,
                ) = coordinator.decision_protocol.make_decision(
                    self.agreements, self.turn, i, task_instruction, input_str, config
                )
                if additional_voting_results:
                    voting_results_per_turn[self.turn] = additional_voting_results
                else:
                    voting_results_per_turn[self.turn] = None

                if self.decision:
                    break

            if coordinator.judge:
                template_filling = TemplateFilling(
                    task_instruction=task_instruction,
                    input_str=input_str,
                    current_draft=self.draft,
                    persona=coordinator.judge.persona,
                    persona_description=coordinator.judge.persona_description,
                    agent_memory=discussion_history,
                )
                self.unique_id, self.turn = coordinator.judge.intervention(self.unique_id, self.turn, memory_ids, template_filling, self.draft, always_intervene=config.judge_always_intervene)

            self.print_messages(coordinator, input_str, task_instruction)

        self.print_messages(
            coordinator,
            input_str,
            task_instruction,
            False,
            solution,
            voting_process_string,
            console,
        )
        return (
            self.draft,
            self.turn,
            self.agreements,
            self.decision,
            voting_results_per_turn,
        )

    def print_messages(
        self,
        coordinator: Coordinator,
        input_str: str,
        task_instruction: str,
        only_current_turn: bool = True,
        solution: str = "",
        voting_process_string: str = "",
        console: Optional[Console] = None,
    ) -> None:
        if console is None:
            console = Console()
        global_memories = [
            memory
            for memory in coordinator.memory
            if memory.turn == self.turn or not only_current_turn
        ]
        if not global_memories:  # if the regenerate judge intervention is triggered, the memory is empty
            return

        max_width = min(console.width, 100)
        discussion_text = Text(
            f"Task instruction: {task_instruction}\n\nInput: {input_str}\n-----------\n"
            + "\n-----------\n".join(
                [
                    f"Agent ({m.persona})({'agreed' if m.agreement else 'disagreed'}): {m.message}"
                    for m in global_memories
                ]
            )
            + f"\n-----------\nDecision Success: {self.decision} \n\nReal Solution: {solution}\nDiscussion solution: {self.draft}"
            + (f"\n\n{voting_process_string}" if voting_process_string else "")
        )
        discussion_text.highlight_regex(r"Agent .*\):", style="bold green")
        discussion_text.highlight_regex(r"Task instruction:", style="bold green")
        discussion_text.highlight_regex(r"Input:", style="bold green")
        discussion_text.highlight_regex(r"Decision Success:", style="bold green")
        discussion_text.highlight_regex(r"Accepted solution:", style="bold green")
        discussion_text.highlight_regex(r"Voting with alteration:", style="bold green")
        discussion_text.highlight_regex(r".* final answer:", style="bold green")
        discussion_text.highlight_regex(r"Facts:", style="bold green")
        discussion_text.highlight_regex(r"####.*", style="bold green")
        for panelist in coordinator.panelists:
            discussion_text.highlight_regex(panelist.persona, style="bold blue")
        panel = Panel(
            discussion_text,
            title=(
                f"Discussion Turn {global_memories[0].turn}"
                if only_current_turn
                else "Discussion"
            ),
            subtitle=f"Decision: {self.decision}",
            expand=False,
            width=max_width,
        )
        console.print(panel)

    @abstractmethod
    def draft_proposer_call(
        self,
        draft_proposer: DraftProposer,
        coordinator: Coordinator,
        agent_index: int,
        memory_ids: list[int],
        template_filling: TemplateFilling,
    ) -> None:
        pass

    @abstractmethod
    def panelist_call(
        self,
        agent: Panelist,
        coordinator: Coordinator,
        agent_index: int,
        memory_ids: list[int],
        template_filling: TemplateFilling,
    ) -> None:
        pass



from typing import Any, Optional

from evaluate import load

from mallm.evaluation.metrics.metric import Metric

# from nltk.translate.bleu_score import sentence_bleu    # TODO: preferred but broken with python 3.12 at commit time


class BLEU(Metric):
    """
    A class to evaluate the BLEU score for text generation tasks.
    """

    _name = "BLEU"

    @staticmethod
    def evaluate(generated_text: str, reference_texts: list[str], dataset_id: Optional[str]) -> dict[str, Any]:
        # Calculate BLEU score
        score = load("bleu", trust_remote_code=True).compute(
            references=[reference_texts], predictions=[generated_text]
        )
        return {"bleu": score["bleu"]}



from collections.abc import Sequence

from mallm.agents.agent import Agent
from mallm.utils.types import Agreement, Memory, TemplateFilling


class Panelist(Agent):
    """
    Represents a Panelist agent, a type of Agent that specializes in participating in discussions.
    It is prompted to either improve the current solution or provide feedback.
    """

    def participate(
        self,
        memories: list[Memory],
        unique_id: int,
        turn: int,
        memory_ids: list[int],
        template_filling: TemplateFilling,
        agents_to_update: Sequence[Agent],
        agreements: list[Agreement],
    ) -> list[Agreement]:
        """
        Either calls feedback() or improve(), depending on whether this panelist is allowed to draft solutions or not.
        """
        if self.drafting_agent:
            _res, memory, agreements = self.improve(
                unique_id=unique_id,
                turn=turn,
                memory_ids=memory_ids,
                template_filling=template_filling,
                agreements=agreements,
            )
        else:
            _res, memory, agreements = self.feedback(
                unique_id=unique_id,
                turn=turn,
                memory_ids=memory_ids,
                template_filling=template_filling,
                agreements=agreements,
            )

        memories.append(memory)
        self.coordinator.update_memories(memories, agents_to_update)
        return agreements



from abc import ABC, abstractmethod
from typing import Any, Optional


class Metric(ABC):
    """
    Abstract class for evaluation metrics in text generation tasks.
    This class provides a template for implementing various evaluation metrics.
    """

    _name = "Metric"

    @staticmethod
    @abstractmethod
    def evaluate(generated_text: str, reference_texts: list[str], dataset_id: Optional[str]) -> dict[str, Any]:
        """
        Evaluate the generated text against a reference text.

        Args:
        generated_text (str): The text generated by the model.
        reference_text (list[str]): The reference text to compare against.

        Returns:
        dict[str, Any]: The evaluation results.
        """

    @property
    def name(self) -> str:
        """
        Return the name of the evaluation metric.

        Returns:
        str: The name of the metric.
        """
        return self._name



from typing import Any, Optional

from nltk import download, word_tokenize
from nltk.translate.meteor_score import meteor_score

from mallm.evaluation.metrics.metric import Metric


class METEOR(Metric):
    """
    A class to evaluate the METEOR score for text generation tasks.
    """

    _name = "METEOR"
    _downloaded = False

    @staticmethod
    def download_and_prepare() -> None:
        download("wordnet")
        download("punkt")
        METEOR._downloaded = True

    @staticmethod
    def evaluate(generated_text: str, reference_texts: list[str], dataset_id: Optional[str]) -> dict[str, Any]:
        if not METEOR._downloaded:
            METEOR().download_and_prepare()
        # Tokenize the input texts
        generated_tokens = word_tokenize(generated_text)
        reference_tokens = [word_tokenize(r) for r in reference_texts]
        # Calculate METEOR score
        score = meteor_score(hypothesis=generated_tokens, references=reference_tokens)
        return {"meteor": score}



# https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
import logging
from typing import ClassVar

from colorama import Fore


class CustomFormatter(logging.Formatter):
    # grey = "\x1b[38;20m"
    # yellow = "\x1b[33;20m"
    # red = "\x1b[31;20m"
    # bold_red = "\x1b[31;1m"
    # reset = "\x1b[0m"
    format_str = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    )
    FORMATS: ClassVar[dict[int, str]] = {
        logging.DEBUG: Fore.BLUE + format_str,
        logging.INFO: Fore.WHITE + format_str,
        logging.WARNING: Fore.YELLOW + format_str,
        logging.ERROR: Fore.LIGHTRED_EX + format_str,
        logging.CRITICAL: Fore.RED + format_str,
    }

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = self.FORMATS.get(record.levelno, Fore.WHITE) + Fore.RESET
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



from abc import ABC, abstractmethod

from mallm.models.Chat import Chat
from mallm.utils.types import InputExample


class PersonaGenerator(ABC):
    @abstractmethod
    def __init__(self, llm: Chat):
        pass

    @abstractmethod
    def generate_persona(
        self,
        task_description: str,
        already_generated_personas: list[dict[str, str]],
        sample: InputExample,
    ) -> dict[str, str]:
        """
        Abstract method to generate a list of persona descriptions based on the given task.

        Parameters:
        task_description (str): Description of the task for which personas are to be generated.
        num_agents (int): Number of persona descriptions to generate.

        Returns:
        list[dict[str, str]]: A list of dictionaries, where each dictionary represents a persona
                               with keys like 'role' and 'persona' (or other relevant descriptors),
                               each mapped to their respective string description.
        """



from typing import Any, Optional

from bert_score import score as bert_score

from mallm.evaluation.metrics.metric import Metric


class BERTScore(Metric):
    """
    A class to evaluate the BERTScore for text generation tasks.
    This includes precision, recall, and F1 score.
    """

    _name = "BERTScore"

    @staticmethod
    def evaluate(generated_text: str, reference_texts: list[str], dataset_id: Optional[str]) -> dict[str, Any]:
        # Calculate BERTScore
        _P, _R, F1 = bert_score(
            cands=[generated_text],
            refs=[reference_texts[0]],
            lang="en",
            model_type="bert-base-uncased",
            num_layers=9,
        )
        return {"bertscore": F1.mean().item()}



from mallm.models.Chat import Chat
from mallm.models.personas.PersonaGenerator import PersonaGenerator
from mallm.utils.types import InputExample


class MockGenerator(PersonaGenerator):
    """
    The MockGenerator class is a placeholder PersonaGenerator that generates a simple persona based on the number of already generated personas..
    """

    def __init__(self, llm: Chat):
        pass

    @staticmethod
    def generate_persona(
        task_description: str,
        already_generated_personas: list[dict[str, str]],
        sample: InputExample,
    ) -> dict[str, str]:
        return {
            "role": f"Participant {len(already_generated_personas) + 1}",
            "description": "generic",
        }



TASK_INSTRUCTIONS: dict[str, str] = {
    "etpc": "Paraphrase the provided text into a single paraphrase by using all paraphrase types.",
    "wmt19_de_en": "Translate the provided text from German to English.",
    "simple_ethical_questions": "Answer the provided question by choosing option A), B), C), or D). Include the letter corresponding to your answer in the solution.",
    "squad_v2": "Answer the following question. If the question is not answerable with the provided information, write '[UNKNOWN]'.",
    "strategyqa": "Answer the following question with A) Yes or B) No. Include the letter corresponding to your answer in the solution.",
    "xsum": "Summarize the provided text into one sentence.",
    "mmlu_pro": "Answer the provided question by choosing one option. Include the letter corresponding to your answer in the solution.",
    "musr": "Solve the mystery. Make sure to always include the full name of the suspect in your solution.",
    "gpqa": "Answer the provided question by choosing option A), B), C), or D). Include the letter corresponding to your answer in the solution.",
    "math_lvl5": "Answer the following math question. Output the number as the final solution.",
    "ifeval": "Follow the instruction.",
    "winogrande": "Correctly fill the blank space _ by choosing option A) or B). Include the letter corresponding to your answer in the solution.",
    "ethics": "Would most people believe this reasonable or unreasonable to say? Answer with A) Reasonable or B) Unreasonable. Include the letter corresponding to your answer in the solution.",
    "aqua_rat": "Answer the following math question. Include the letter corresponding to your answer in the solution."
}



import json
import logging
from pathlib import Path
from typing import Any, Optional

import fire
import json_repair
from tqdm import tqdm

from mallm.evaluation.metrics.bertscore import BERTScore
from mallm.evaluation.metrics.bleu import BLEU
from mallm.evaluation.metrics.ifeval import IFEval
from mallm.evaluation.metrics.meteor import METEOR
from mallm.evaluation.metrics.metric import Metric
from mallm.evaluation.metrics.qa import (
    AnswerabilityBoolean,
    IncludesAnswer,
    MultiChoiceBoolean,
    SquadScore,
)
from mallm.evaluation.metrics.rouge import ROUGE
from mallm.evaluation.plotting.plots import create_plots_for_path

ALL_METRICS = [
    AnswerabilityBoolean(),
    BERTScore(),
    BLEU(),
    METEOR(),
    MultiChoiceBoolean(),
    ROUGE(),
    SquadScore(),
    IncludesAnswer(),
    IFEval(),
]

logger = logging.getLogger("mallm")


class Evaluator:
    def __init__(
        self,
        input_file_path: str,
        output_dir_path: Optional[str] = None,
        metrics: Optional[list[str]] = None,
        extensive: bool = False,
    ) -> None:
        self.input_file_path = Path(input_file_path)
        if not self.input_file_path.exists() or not self.input_file_path.is_file():
            raise FileNotFoundError(f"Input file not found: {self.input_file_path}")

        if output_dir_path:
            output_dir = Path(output_dir_path)
            output_file_path = output_dir
        else:
            output_file_path = self.input_file_path

        self.stats_file_path = output_file_path.with_name(
            output_file_path.stem + "-stats.json"
        )
        self.eval_file_path = output_file_path.with_name(
            output_file_path.stem + "-eval.json"
        )

        self.data = self._load_data()
        self.metrics = self._initialize_metrics(metrics)
        self.extensive = extensive

    def _load_data(self) -> list[dict[str, Any]]:
        data_str = Path(self.input_file_path).read_text()
        data: list[dict[str, Any]] = json_repair.repair_json(
            data_str, return_objects=True
        )
        return data

    @staticmethod
    def _initialize_metrics(metrics: Optional[list[str]]) -> list[Any]:
        if metrics is None:
            metrics = ["multichoice"]

        metrics = [m.lower() for m in metrics]
        selected_metrics = [
            metric_class
            for metric_class in ALL_METRICS
            if metric_class.name.lower() in metrics
        ]
        if len(selected_metrics) != len(metrics):
            logger.warning(f"Some metrics not found in {metrics}")
        if not selected_metrics:
            raise ValueError(f"No metrics found for {metrics}")
        logger.info(f"Metrics to calculate: {[m.name for m in selected_metrics]}")
        return selected_metrics

    @staticmethod
    def calculate_score(
        answer: str,
        references: list[str],
        metric: Metric,
        dataset_id: Optional[str] = None,
    ) -> dict[str, Any]:
        return next(iter(metric.evaluate(answer, references, dataset_id).values()))

    def calculate_scores(
        self,
        answer: str,
        references: list[str],
        metric_alteration: str = "",
        dataset_id: Optional[str] = None,
    ) -> dict[str, Any]:
        metrics = []
        if references:
            metrics.extend(self.metrics)
        else:
            if any(metric.name == "answerability" for metric in self.metrics):
                metrics.append(AnswerabilityBoolean())
            if any(metric.name == "squad" for metric in self.metrics):
                metrics.append(SquadScore())
            if any(metric.name == "IFEval" for metric in self.metrics):
                metrics.append(IFEval())
        if not metrics:
            logger.warning(f"No metrics to evaluate against references {references}.")
            return {}

        return {
            f"{k}{f'-{metric_alteration}' if metric_alteration else ''}": v
            for metric in metrics
            for k, v in metric.evaluate(answer, references, dataset_id).items()
        }

    def add_scores(self) -> None:
        for item in tqdm(
            self.data, desc=f"Calculating scores of {self.input_file_path}: "
        ):
            main_answer = item.get("finalAnswer", "")
            references = item.get("references", [])
            dataset_id = item.get("datasetId", None)
            if main_answer:
                item["scores"] = self.calculate_scores(
                    main_answer, references, "", dataset_id
                )

            votes_each_turn = item.get("votesEachTurn", None)
            if votes_each_turn:
                alterations: dict[str, Any] = votes_each_turn[
                    max(votes_each_turn.keys())
                ].get("alterations", "")
                if alterations and len(alterations) >= 1:
                    for alteration in list(alterations.keys()):
                        answer = alterations[alteration].get("final_answer", "")
                        if answer and "scores" not in item:
                            item["scores"] = self.calculate_scores(
                                answer, references, alteration, dataset_id
                            )
                        elif answer:
                            item["scores"].update(
                                self.calculate_scores(answer, references, alteration, dataset_id)
                            )

            challenged_answers: Any = item.get("challengedAnswers", None)
            if challenged_answers:
                if "scores" not in item:
                    continue
                if "correct" not in item["scores"] and "f1" not in item["scores"]:
                    continue
                if challenged_answers["challenged_answers"]:
                    self.analyze_challenged_answers(
                        "normal",
                        challenged_answers["challenged_answers"],
                        item,
                        references,
                        item["scores"],
                        dataset_id
                    )
                if challenged_answers["challenged_answers_wrong"]:
                    self.analyze_challenged_answers(
                        "wrong",
                        challenged_answers["challenged_answers_wrong"],
                        item,
                        references,
                        self.calculate_scores(
                            challenged_answers["wrong_answer"], references, "", dataset_id
                        ),
                        dataset_id
                    )
                if challenged_answers["challenged_answers_irrelevant"]:
                    self.analyze_challenged_answers(
                        "irrelevant",
                        challenged_answers["challenged_answers_irrelevant"],
                        item,
                        references,
                        self.calculate_scores(
                            challenged_answers["irrelevant_answer"], references, "", dataset_id
                        ),
                        dataset_id
                    )
                if challenged_answers["challenged_answers_history"]:
                    self.analyze_challenged_answers(
                        "history",
                        challenged_answers["challenged_answers_history"],
                        item,
                        references,
                        item["scores"],
                        dataset_id
                    )
                if challenged_answers["challenged_answers_additional_information"]:
                    self.analyze_challenged_answers(
                        "information",
                        challenged_answers["challenged_answers_additional_information"],
                        item,
                        references,
                        item["scores"],
                        dataset_id
                    )

    def analyze_challenged_answers(
        self,
        name: str,
        challenged_answers: dict[str, Optional[str]],
        item: Any,
        references: list[str],
        previous_score: Any,
        dataset_id: Optional[str] = None,
    ) -> None:
        new_answer = {
            f"{name}_no_challenge": True,
            f"{name}_challenge_failed": False,
            f"{name}_challenge_higher": False,
            f"{name}_challenge_lower": False,
            f"{name}_challenge_same": False,
        }
        previous_score = (
            previous_score.get("f1")
            if previous_score.get("f1", None) is not None
            else previous_score.get("correct", None)
        )

        answer = next(iter(challenged_answers.values()))
        if answer:
            score = self.calculate_scores(answer, references, "", dataset_id)
            current_score = (
                score.get("f1")
                if score.get("f1", None) is not None
                else score.get("correct", None)
            )
            if current_score is None or previous_score is None:
                new_answer[f"{name}_challenge_failed"] = True
            elif current_score > previous_score:
                new_answer[f"{name}_challenge_higher"] = True
            elif current_score < previous_score:
                new_answer[f"{name}_challenge_lower"] = True
            elif current_score == previous_score:
                new_answer[f"{name}_challenge_same"] = True
            new_answer[f"{name}_no_challenge"] = False
        item["scores"].update(new_answer)

    def add_scores_extensive(self) -> None:
        for item in tqdm(self.data, desc="Extensive scores: "):
            references = item.get("references", [])
            dataset_id = item.get("datasetId", None)
            votes_each_turn = item.get("votesEachTurn", None)
            alterations: dict[str, Any] = votes_each_turn[
                max(votes_each_turn.keys())
            ].get("alterations", None)
            for mem in item.get("globalMemory", []):
                solution = mem.get("solution", "")
                if solution and alterations:
                    for alteration in list(alterations.keys()):
                        if "scores" not in mem:
                            mem["scores"] = self.calculate_scores(
                                solution, references, alteration, dataset_id
                            )
                        else:
                            mem["scores"].update(
                                self.calculate_scores(solution, references, alteration, dataset_id)
                            )
                elif solution:
                    score = self.calculate_scores(solution, references, "", dataset_id)
                    mem["scores"] = score

            if votes_each_turn:
                for turn in votes_each_turn:
                    if alterations:
                        for alteration in list(alterations.keys()):
                            item["votesEachTurn"][turn]["alterations"][alteration][
                                "score"
                            ] = self.calculate_scores(
                                item["votesEachTurn"][turn]["alterations"][alteration][
                                    "final_answer"
                                ],
                                references,
                                alteration,
                                dataset_id
                            )

    def calculate_statistics(self) -> dict[str, Any]:
        reported_metrics = set()
        for item in self.data:
            if "scores" in item:
                reported_metrics.update(item["scores"].keys())

        if not reported_metrics:
            logger.error("No elements with scores found in the data.")
            raise Exception("No elements with scores found in the data.")

        logger.info(f"Reported metrics: {reported_metrics}")

        stats = {}
        for metric in reported_metrics:
            scores = [item.get("scores", {}).get(metric) for item in self.data]
            scores = [
                score
                for score in scores
                if isinstance(score, (int, float)) and score is not None
            ]

            if not scores:
                logger.warning(f"No numeric scores found for {metric}")
                continue

            average_score = sum(scores) / len(scores)
            if len(scores) > 1:
                std_dev_score = (
                    sum((score - average_score) ** 2 for score in scores)
                    / (len(scores) - 1)
                ) ** 0.5
            else:
                std_dev_score = 0

            avg_scores_per_turn = {}
            if self.extensive:
                for item in self.data:
                    for mem in item.get("globalMemory", []):
                        turn = mem.get("turn", 0)
                        if turn not in avg_scores_per_turn:
                            avg_scores_per_turn[turn] = float(0)
                        turn_score = mem.get("scores", {}).get(metric, 0)
                        if turn_score:
                            avg_scores_per_turn[turn] += turn_score

                max_turns = max(item.get("turns", 0) for item in self.data)
                for turn in range(max_turns + 1)[1:]:
                    num_scores = sum(
                        1
                        for item in self.data
                        for mem in item.get("globalMemory", [])
                        if mem.get("turn", 0) == turn
                        and metric in mem.get("scores", {})
                    )
                    if num_scores == 0:
                        num_scores = 1
                    avg_scores_per_turn[turn] /= num_scores
                    avg_scores_per_turn[turn] = round(avg_scores_per_turn[turn], 4)

            stats[metric] = {
                "dataSize": len(self.data),
                "sampleSize": len(scores),
                "scores": scores,
                "averageScore": round(average_score, 4),
                "stdDevScore": round(std_dev_score, 4),
                "averageScoresPerTurnAggregated": avg_scores_per_turn,
            }

        return stats

    def save_results(self, stats: dict[str, Any]) -> None:
        self.stats_file_path.write_text(json.dumps(stats, indent=4))
        self.eval_file_path.write_text(json.dumps(self.data, indent=4))
        logger.info(f"Statistics saved to {self.stats_file_path}")
        logger.info(f"Eval saved to {self.eval_file_path}")

    def process(self) -> None:
        self.add_scores()
        if self.extensive:
            self.add_scores_extensive()
        stats = self.calculate_statistics()
        self.save_results(stats)


def batch_process_dir_path(
    input_dir_path: str,
    output_dir_path: Optional[str] = None,
    metrics: Optional[list[str]] = None,
    extensive: bool = False,
) -> None:
    input_path = Path(input_dir_path)
    output_path = Path(output_dir_path) if output_dir_path else input_path
    if not output_path.exists() or not output_path.is_dir():
        print(f"Creating output directory: {output_path}")
        output_path.mkdir(parents=True, exist_ok=True)
    if not input_path.exists() or not input_path.is_dir():
        raise FileNotFoundError(f"Input directory not found: {input_path}")
    print(f"Processing files in {input_path} and saving results to {output_path}")
    print(f"Metrics to calculate: {metrics}")
    print(f"Files to process: {len(list(input_path.glob('*.json')))}")
    files = list(input_path.glob("*.json"))
    for file in files:
        if file.stem.endswith(("-eval", "-stats")):
            continue

        output_file = output_path / file.name
        logger.info(f"Processing {file}")

        evaluator = Evaluator(str(file), str(output_file), metrics, extensive)
        evaluator.process()

    logger.info("Batch processing completed.")
    logger.info("Creating plots...")
    plot_path = str(output_path).removesuffix("/")
    create_plots_for_path(plot_path, plot_path)
    logger.info("Plots created.")


def run_evaluator(
    input_json_file_path: str,
    output_dir_path: Optional[str] = None,
    metrics: Optional[list[str]] = None,
    extensive: bool = False,
) -> None:
    if Path(input_json_file_path).is_dir():
        batch_process_dir_path(
            input_json_file_path, output_dir_path, metrics, extensive
        )
    else:
        evaluator = Evaluator(input_json_file_path, output_dir_path, metrics, extensive)
        evaluator.process()


def main() -> None:
    fire.Fire(run_evaluator)


if __name__ == "__main__":
    main()



from enum import Enum


class DecisionAlteration(Enum):
    PUBLIC = "public"
    FACTS = "facts"
    HISTORY = "history"
    CONFIDENCE = "confidence"
    CONFIDENCE_LOG_PROBS = "confidence_log_probs"
    CONFIDENCE_PROMPTED = "confidence_prompted"
    CONFIDENCE_CONSISTENCY = "confidence_consistency"
    ANONYMOUS = "anonymous"



import logging
from typing import Any, Optional

from mallm.agents.panelist import Panelist
from mallm.decision_protocol.protocol import DecisionProtocol
from mallm.models.discussion.ResponseGenerator import ResponseGenerator
from mallm.utils.config import Config
from mallm.utils.enums import DecisionAlteration
from mallm.utils.types import Agreement, VotingResult, VotingResultList, WorkerFunctions

logger = logging.getLogger("mallm")


class RankedVoting(DecisionProtocol):
    """
    The Ranked Voting decision protocol allows panelists to rank their preferences among a set of answers after a certain number of turns.
    """

    _name = "ranked_voting"

    def __init__(
        self,
        panelists: list[Panelist],
        num_neutral_agents: int,
        worker_functions: WorkerFunctions,
        vote_turn: int = 3,
    ) -> None:
        super().__init__(panelists, num_neutral_agents, worker_functions)
        self.vote_turn = vote_turn

    def make_decision(
        self,
        agreements: list[Agreement],
        turn: int,
        agent_index: int,
        task: str,
        question: str,
        config: Config,
    ) -> tuple[str, bool, list[Agreement], str, Optional[VotingResultList]]:
        if len(agreements) > self.total_agents:
            agreements = agreements[-self.total_agents :]

        if turn < self.vote_turn or agent_index != self.total_agents - 1:
            return "", False, agreements, "", None

        final_answers_with_confidence, voting_process_string = (
            self.generate_final_answers(agreements, question, task)
        )

        decision, final_answer, results, voting_process_string = (
            self.vote_with_alterations(
                final_answers_with_confidence,
                question,
                task,
                voting_process_string,
                self._name,
                ResponseGenerator.generate_ranking_prompt,
                config.voting_protocols_with_alterations,
            )
        )
        return (
            final_answer,
            decision,
            agreements,
            voting_process_string,
            results,
        )

    def process_results(
        self,
        all_votes: dict[str, VotingResult],
        alteration: DecisionAlteration,
        final_answers: list[str],
        votes: list[list[int]],
    ) -> dict[str, VotingResult]:
        # Calculate the score for each answer based on the rankings
        scores = [0] * len(final_answers)
        for ranking_list in votes:
            for rank, idx in enumerate(ranking_list):
                scores[idx] += (
                    min(5, self.total_agents) - rank
                )  # Score 5 for the 1st rank, 4 for the 2nd, etc.

        # Find the answer with the highest score
        highest_score = max(scores)
        index = scores.index(highest_score)
        best_answers = [
            final_answers[i] for i, score in enumerate(scores) if score == highest_score
        ]

        # If there's a tie, pick the first answer among the best
        # If all panelists agree on the best answer finished else go for another round
        if len(best_answers) == 1:
            all_votes[alteration.value] = VotingResult(
                votes=votes,
                most_voted=index,
                final_answer=final_answers[index],
                agreed=True,
            )
            logger.info(
                f"Selected answer from agent {self.panelists[index].short_id} with {highest_score} points"
            )
        else:
            all_votes[alteration.value] = VotingResult(
                votes=votes,
                most_voted=-1,
                final_answer="",
                agreed=False,
            )
            logger.info("There was a tie. Going for another round of voting.")
        return all_votes

    def process_votes(
        self,
        final_answers: list[str],
        panelist: Panelist,
        vote_str: str,
        vote: Any,
        voting_process_string: str,
    ) -> tuple[str, Any, bool, str]:
        success = False
        # Split the ranking and convert to a list of integers
        ranking_list = list(map(int, vote_str.strip().split()))
        if (
            all(0 <= rank < len(final_answers) for rank in ranking_list)
            and len(ranking_list) <= 5
        ):
            vote.append(ranking_list)
            logger.info(
                f"{panelist.persona} ranked answers: {[self.panelists[a].persona for a in ranking_list]}"
            )
            voting_process_string += f"{panelist.persona} ranked answers: {[self.panelists[a].persona for a in ranking_list]}\n"
            success = True
        return vote_str, vote, success, voting_process_string



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from mallm.models.Chat import Chat
from mallm.utils.types import Response, TemplateFilling

if TYPE_CHECKING:
    from mallm.agents.panelist import Panelist


class ResponseGenerator(ABC):
    @abstractmethod
    def __init__(self, llm: Chat):
        pass

    @abstractmethod
    def generate_response(
        self,
        current_prompt: list[dict[str, str]],
        task_instruction: str,
        input_str: str,
        chain_of_thought: bool,
        agreement: Optional[bool],
        baseline: bool,
        drafting: bool,
    ) -> Response:
        """
        Abstract method to generate an agents response to a discussion.

        Parameters:
        current_prompt (list[dict[str, str]]): The prompt for the LM.
        chain_of_thought (bool): Whether to use Zero-Shot CoT.
        agreement (Optional[bool]): the agreement if already computed.
        baseline (bool): Whether use the prompt for the baseline, without discussion.
        drafting (bool): Whether the response should be drafting a new solution.

        Returns:
        Response: An object with the attributes "agreement", "message", and "solution".
        """

    @abstractmethod
    def generate_baseline(
        self, task_instruction: str, input_str: str, chain_of_thought: bool
    ) -> Response:
        """
        Abstract method to generate the response of a single LM as a baseline.

        Parameters:
        task_instruction (str): The instruction for the task and appended context.
        input_str (str): The input for the task.
        chain_of_thought (bool): Whether to use Zero-Shot CoT.

        Returns:
        Response: An object with the attributes "agreement", "message", and "solution".
        """

    @abstractmethod
    def generate_ablation(
        self,
        task_instruction: str,
        input_str: str,
        current_solution: str,
        chain_of_thought: bool,
    ) -> Response:
        """
        Abstract method to generate the response of a single LM that iteratively improves on the current solution.

        Parameters:
        task_instruction (str): The instruction for the task and appended context.
        input_str (str): The input for the task.
        current_solution: (str): The current solution to improve.
        chain_of_thought (bool): Whether to use Zero-Shot CoT.

        Returns:
        Response: An object with the attributes "agreement", "message", and "solution".
        """

    @abstractmethod
    def generate_feedback(
        self, data: TemplateFilling, chain_of_thought: bool
    ) -> Response:
        """
        Abstract method to generate feedback to improve a draft for a discussion.

        Parameters:
        data (TemplateFilling): The fields used for prompting the LM.
        chain_of_thought (bool): Whether to use Zero-Shot CoT.

        Returns:
        Response: An object with the attributes "agreement", "message", and "solution".
        """

    @abstractmethod
    def generate_improve(
        self, data: TemplateFilling, chain_of_thought: bool
    ) -> Response:
        """
        Abstract method to generate feedback and an improved solution for a discussion.

        Parameters:
        data (TemplateFilling): The fields used for prompting the LM.
        chain_of_thought (bool): Whether to use Zero-Shot CoT.

        Returns:
        Response: An object with the attributes "agreement", "message", and "solution".
        """

    @abstractmethod
    def generate_draft(self, data: TemplateFilling, chain_of_thought: bool) -> Response:
        """
        Abstract method to generate a draft for a discussion.

        Parameters:
        data (TemplateFilling): The fields used for prompting the LM.
        chain_of_thought (bool): Whether to use Zero-Shot CoT.

        Returns:
        Response: An object with the attributes "agreement", "message", and "solution".
        """

    @staticmethod
    def get_filled_template(data: TemplateFilling) -> list[dict[str, str]]:
        prompt_str = f"""
Task: {data.task_instruction}
Input: {data.input_str}
Your role: {data.persona} ({data.persona_description})
Current Solution: {data.current_draft}
"""  # input has context appended

        appendix = ""
        if data.current_draft is None:
            appendix += (
                "\nNobody proposed a solution yet. Please provide the first one."
            )
        if data.agent_memory is not None and data.agent_memory != []:
            appendix += "\nThis is the discussion to the current point: \n"
        prompt = [
            {
                "role": "user",
                "content": prompt_str + appendix,
            }
        ]
        if data.agent_memory is not None and data.agent_memory != []:
            prompt += data.agent_memory
        return prompt

    @staticmethod
    def get_filled_template_slim(data: TemplateFilling) -> list[dict[str, str]]:
        prompt_str = f"""
Task: {data.task_instruction}
Input: {data.input_str}
Your role: {data.persona} ({data.persona_description})
Current Solution: {data.current_draft}
"""  # input has context appended

        appendix = ""
        if data.current_draft is None:
            appendix += (
                "\nNobody proposed a solution yet. Please provide the first one."
            )
        if data.agent_memory is not None and data.agent_memory != []:
            appendix += "\nThis is the discussion to the current point: \n"
        prompt = [
            {
                "role": "user",
                "content": prompt_str + appendix,
            }
        ]
        if data.agent_memory is not None and data.agent_memory != []:
            prompt += data.agent_memory
        return prompt

    @staticmethod
    def extract_agreement(res: str, drafting: bool) -> Optional[bool]:
        """
        Determines whether a string given by an agent means an agreement or disagreement.
        Returns bool
        """
        if drafting:
            return None
        return "agree" in res.lower() and "disagree" not in res.lower()

    @staticmethod
    def generate_final_answer_prompt(
        input_sample: str,
        task: str,
        previous_answer: Optional[str],
        persona: Optional[str] = None,
        persona_description: Optional[str] = None,
    ) -> list[dict[str, str]]:

        prompt = []
        if persona:
            prompt.append(
                {
                    "role": "system",
                    "content": f"Your role: {persona} ({persona_description})",
                },
            )
        prompt.extend(
            [
                {
                    "role": "system",
                    "content": "You are tasked with creating a final solution based on the given input and your previous response.",
                },
                {
                    "role": "user",
                    "content": f"Task: {task}\nInput: {input_sample}\nYour previous response: {previous_answer}",
                },
                {
                    "role": "user",
                    "content": "Extract the final solution to the task from the provided text. Remove statements of agreement, disagreement, and explanations. Do not modify the text. Do not output any text besides the solution. If there is no solution provided, just copy the previous response.",
                },
            ]
        )
        return prompt

    @staticmethod
    def generate_answer_confidence_prompt(
        panelist: Panelist,
        question: str,
        task: str,
        final_answer: str,
    ) -> list[dict[str, str]]:
        prompts = [
            {
                "role": "system",
                "content": f"Your role: {panelist.persona} ({panelist.persona_description})",
            }
        ]
        discussion_history = panelist.get_discussion_history()[0]
        if discussion_history:
            prompts.append(
                {
                    "role": "user",
                    "content": "Here is the discussion history to help you make a decision:",
                }
            )
            prompts.extend(discussion_history)
        prompts.append(
            {
                "role": "user",
                "content": f"The task is: {task}. The question is: {question}. This is the final answer you provided: '{final_answer}'. Based on this information, please generate a confidence score between 0 and 100 for the final answer. Be critical and only answer with the number.",
            }
        )
        return prompts

    @staticmethod
    def generate_challenge_prompt(
        panelist: Panelist,
        question: str,
        task: str,
        final_answer: str,
        history: bool = False,
        facts: Optional[str] = None,
    ) -> list[dict[str, str]]:
        agent_history = panelist.get_discussion_history()[0] if history else []
        prompts = [
            {
                "role": "system",
                "content": f"You are a participant in a group discussion. Your role: {panelist.persona} ({panelist.persona_description})",
            }
        ]
        if history and agent_history:
            prompts.append(
                {
                    "role": "system",
                    "content": "This is the discussion to the current point:",
                }
            )
            prompts.extend(agent_history)
        if facts:
            prompts.append(
                {
                    "role": "system",
                    "content": f"Here is some helpful additional information to improve your answer quality: {facts}",
                }
            )
        prompts.append(
            {
                "role": "user",
                "content": (
                    f"The task is: {task}. The question is: {question}. "
                    f"This is the final answer generated by the discussion: '{final_answer}'. "
                    "Please critically evaluate this answer. If you agree with the final answer, respond with the exact word 'AGREE' to confirm. If you do not agree, respond with the exact word 'DISAGREE' to challenge the answer."
                ),
            },
        )
        return prompts

    @staticmethod
    def generate_challenge_new_answer_prompt(
        panelist: Panelist,
        question: str,
        task: str,
        final_answer: str,
        history: bool = False,
        facts: Optional[str] = None,
    ) -> list[dict[str, str]]:
        agent_history = panelist.get_discussion_history()[0] if history else []
        prompts = [
            {
                "role": "system",
                "content": f"You are a participant in a group discussion. Your role: {panelist.persona} ({panelist.persona_description})",
            }
        ]
        if history and agent_history:
            prompts.append(
                {
                    "role": "system",
                    "content": "This is the discussion to the current point:",
                }
            )
            prompts.extend(agent_history)
        if facts:
            prompts.append(
                {
                    "role": "system",
                    "content": f"Here is some helpful additional information to improve your answer quality: {facts}",
                }
            )
        prompts.append(
            {
                "role": "user",
                "content": (
                    f"The task is: {task}. The question is: {question}. "
                    f"This is the final answer generated by the discussion: '{final_answer}'. "
                    "You dont agree with the final answer. Please provide a new answer to the question. Include the letter corresponding to your answer in the solution."
                ),
            },
        )
        return prompts

    @staticmethod
    def voting_base_prompt(
        voting_message: str,
        panelist: Panelist,
        panelists: list[Panelist],
        task: str,
        question: str,
        solutions: list[str],
        additional_context: Optional[str] = None,
        anonymous: bool = True,
        confidence: Optional[list[int]] = None,
        history: bool = False,
    ) -> list[dict[str, str]]:
        prompts = [
            {
                "role": "system",
                "content": f"Your role: {panelist.persona} ({panelist.persona_description})",
            }
        ]
        if history:
            discussion_history = panelist.get_discussion_history()[0]
            if discussion_history:
                prompts.append(
                    {
                        "role": "user",
                        "content": "Here is the discussion history to help you make a decision:",
                    }
                )
                prompts.extend(discussion_history)
        additional_context_str = (
            f"\nAdditional Context: {additional_context}" if additional_context else ""
        )
        content_str = (
            f"{voting_message}\n"
            f"Task: {task}\n"
            f"Question: {question}"
            f"{additional_context_str}\n\n"
            "Here are the possible solutions:"
        )
        prompts.append(
            {
                "role": "user",
                "content": content_str,
            }
        )
        for i, solution in enumerate(solutions):
            confidence_str = (
                ""
                if confidence is None
                else f"\n\n(Confidence: {round(confidence[i])} %)"
            )
            prompts.append(
                {
                    "role": "user",
                    "content": f"Solution {i if anonymous else panelists[i].persona}: {solution}{confidence_str}",
                }
            )
        return prompts

    @staticmethod
    def generate_voting_prompt(
        panelist: Panelist,
        panelists: list[Panelist],
        task: str,
        question: str,
        solutions: list[str],
        additional_context: Optional[str] = None,
        anonymous: bool = True,
        confidence: Optional[list[int]] = None,
        history: bool = False,
    ) -> list[dict[str, str]]:
        prompts = ResponseGenerator.voting_base_prompt(
            "You are tasked with voting for the best solution from the list provided below based on the given task.",
            panelist,
            panelists,
            task,
            question,
            solutions,
            additional_context,
            anonymous,
            confidence,
            history,
        )

        prompts.append(
            {
                "role": "user",
                "content": "Based on the above solutions, please provide the number of the solution you are voting for. Answer only with the number.",
            }
        )

        return prompts

    @staticmethod
    def generate_approval_voting_prompt(
        panelist: Panelist,
        panelists: list[Panelist],
        task: str,
        question: str,
        solutions: list[str],
        additional_context: Optional[str] = None,
        anonymous: bool = True,
        confidence: Optional[list[int]] = None,
        history: bool = False,
    ) -> list[dict[str, str]]:
        prompts = ResponseGenerator.voting_base_prompt(
            "You are tasked with approving any number of solutions from the list provided below based on the given task.",
            panelist,
            panelists,
            task,
            question,
            solutions,
            additional_context,
            anonymous,
            confidence,
            history,
        )

        prompts.append(
            {
                "role": "user",
                "content": "Based on the above solutions, please provide the numbers of the solutions you are approving, separated by commas. Answer only with the numbers.",
            }
        )

        return prompts

    @staticmethod
    def generate_cumulative_voting_prompt(
        panelist: Panelist,
        panelists: list[Panelist],
        task: str,
        question: str,
        solutions: list[str],
        additional_context: Optional[str] = None,
        anonymous: bool = True,
        confidence: Optional[list[int]] = None,
        history: bool = False,
    ) -> list[dict[str, str]]:
        prompts = ResponseGenerator.voting_base_prompt(
            "You are tasked with distributing 10 points among the provided solutions based on the given task.",
            panelist,
            panelists,
            task,
            question,
            solutions,
            additional_context,
            anonymous,
            confidence,
            history,
        )

        prompts.append(
            {
                "role": "user",
                "content": "Based on the above solutions, please distribute 10 points among the solutions. Provide your points allocation as a JSON dictionary where keys are solution numbers (as int) and values are the points. The total points should sum up to 10. Answer only with the JSON dictionary.",
            }
        )

        return prompts

    @staticmethod
    def generate_ranking_prompt(
        panelist: Panelist,
        panelists: list[Panelist],
        task: str,
        question: str,
        solutions: list[str],
        additional_context: Optional[str] = None,
        anonymous: bool = True,
        confidence: Optional[list[int]] = None,
        history: bool = False,
    ) -> list[dict[str, str]]:
        prompts = ResponseGenerator.voting_base_prompt(
            "You are tasked with ranking the solutions from the most preferred to the least preferred based on the given task.",
            panelist,
            panelists,
            task,
            question,
            solutions,
            additional_context,
            anonymous,
            confidence,
            history,
        )

        prompts.append(
            {
                "role": "user",
                "content": "Based on the above solutions, please provide the rankings of the solutions separated by spaces. Example: '0 2 1' if you prefer Solution 0 the most, then Solution 2, and finally Solution 1. Provide up to 5 rankings. Only answer with the rankings.",
            }
        )

        return prompts

    @staticmethod
    def generate_summary_prompt(
        panelist: Panelist,
        panelists: list[Panelist],
        task: str,
        question: str,
        solutions: list[str],
        additional_context: Optional[str] = None,
        anonymous: bool = True,
        confidence: Optional[list[int]] = None,
        history: bool = False,
    ) -> list[dict[str, str]]:
        prompts = []

        # Add discussion history if available
        if history:
            discussion_history = panelist.get_discussion_history()[0]
            if discussion_history:
                prompts.append(
                    {
                        "role": "user",
                        "content": "Here is the discussion history to help you make a decision:",
                    }
                )
                prompts.extend(discussion_history)

        # Prepare the main content for the summary request
        additional_context_str = (
            f"\nAdditional Context: {additional_context}" if additional_context else ""
        )

        content_str = (
            f"Task: {task}\n"
            f"Question: {question}"
            f"{additional_context_str}\n\n"
            "Please provide a summary of the following solutions and combine them in a single answer to solve the task. Only answer with the solution:"
        )

        # Add each solution to the content string
        for i, solution in enumerate(solutions):
            confidence_str = (
                "" if confidence is None else f" (Confidence: {round(confidence[i])}%)"
            )
            panelist_label = f"Solution {i}" if anonymous else f"{panelists[i].persona}"
            content_str += f"\n\n{panelist_label}: {solution}{confidence_str}"

        # Append the final content as a user message
        prompts.append(
            {
                "role": "user",
                "content": content_str,
            }
        )

        # Return the prompts list
        return prompts

    @staticmethod
    def generate_wrong_answer_prompt(task: str, question: str) -> list[dict[str, str]]:
        return [
            {
                "role": "system",
                "content": "You are tasked with providing an incorrect or wrong response to the given task and question.",
            },
            {
                "role": "user",
                "content": f"Task: {task}\nQuestion: {question}. Please provide an answer that is deliberately incorrect or inaccurate. Only answer with the incorrect response.",
            },
        ]

    @staticmethod
    def generate_irrelevant_answer_prompt(question: str) -> list[dict[str, str]]:
        return [
            {
                "role": "system",
                "content": "You are tasked with providing a completely unrelated response to the given question.",
            },
            {
                "role": "user",
                "content": f"Question: {question} \n\nPlease provide an answer that is irrelevant to the question. Only answer with the irrelevant response.",
            },
        ]

    @abstractmethod
    def generate_policy_intervention(self, data: TemplateFilling, provide_labels: bool = True) -> Response:
        """
        Generates a policy feedback for the given data. The feedback should give tips on how to improve the discussion.

        Parameters:
        data (TemplateFilling): The data to generate a policy feedback for.
        provide_labels (bool): Whether to provide labels for the feedback.

        Returns:
        Response: An object with the attributes "agreement", "message", and "solution".
        """

    @abstractmethod
    def generate_judgement(self, data: TemplateFilling, answer_before: str, answer_after: str) -> Response:
        """
        Generates the judgement for two solutions, indicating whether discussion is on track or not.

        Parameters:
        data (TemplateFilling): The data to generate a judgement for.
        answer_before (str): The first solution.
        answer_after (str): The second solution.

        Returns:
        Response: An object with the attributes "agreement", "message", and "solution".
        """
